
/* 
 * File: "CoxeterExtensions_by_EM_InvolPoset.cpp"
 *
 * Author: Eric Marberg
 * Version: June 2013
 * --------------------------------------------------------------
 * Source file for "CoxeterExtensions_by_EM_InvolPoset.h"
 *
 */




#include "CoxeterExtensions_by_EM_InvolPoset.h"




/*
 * Function: Delete_3D_PolynomialArray
 * ---------------------------------------------------------
 * Frees memory used for size-by-size SafePolynomial**
 *
 */
void Delete_3D_PolynomialArray(SafePolynomial*** arr, int size)
{
	for(int i=0; i<size; i++) 
	{
		SafePolynomial::Delete_2D_PolynomialArray(arr[i],size);
	}
	delete[] arr;
}



/*
 * Function: Delete_2D_VectorArray
 * ---------------------------------------------------------
 * Frees memory used for size-by-size SafePolynomial**
 *
 */
void Delete_2D_VectorArray(InvolIndexPol_Map** arr, int size)
{	
	for(int i=0; i < size; i++) delete[] arr[i];
	delete[] arr;
}



/*
 * Function: GetPrintingSetting
 * ---------------------------------------------------------
 * Queries the user about how many command line updates to print
 * during calculations.
 *
 */
bool GetPrintingSetting()
{
	cout << endl << "verbose printing (y/n) : ";
	
	char answer[2];
	cin.getline(answer,2);
	
	bool verbose_printing_on = answer[0] == 'y' || answer[0] == 'Y';
	cout << endl << "(Setting verbose printing to " << (verbose_printing_on ? "on" : "off") << ")" << endl << endl;
	
	return verbose_printing_on;
}





/*
 * Function: TwistedInvolutionsPoset constructor
 * ---------------------------------------------------------
 *
 */
TwistedInvolutionsPoset::TwistedInvolutionsPoset()
{	
	// Get current Coxeter group, already checked to be finite
	W = commands::currentGroup();
	
	// Gets the definition of the involution * from the user
	GetGroupAutomorphism();
	
	// Asks user how many command line updates to display
	verbose_printing_on = GetPrintingSetting();
	
	// Computes the list of twisted involutions w and the values of sws^* for s in S
	std::map<CoxNbr,int> index_map;	
	GetInvolutions(index_map);	
	ComputeActions(index_map);
	
	// Writes to a file the reduced expressions for I_* and W in order of index
	RecordIndexing();
	
	// Sets up all data used to work with Bruhat order on I_* and W
	SetupBruhatOrderData();
}



/*
 * Function: TwistedInvolutionsPoset destructor
 * ---------------------------------------------------------
 * Frees memory associated with TwistedInvolutionsPoset
 *
 */
TwistedInvolutionsPoset::~TwistedInvolutionsPoset() 
{
	for(int i=0; i<size(); i++) 
	{
		delete[] elem_actions[i];
		delete[] bruhat_order[i];
	}
	
	delete[] elem_lengths;
	delete[] bruhat_ascents;
	
	for(int i=0; i<W->rank(); i++)
	{
		for(int j=0; j<size(); j++) delete[] bruhat_ascents_by_descents[i][j];
		delete[] bruhat_ascents_by_descents[i];
	}
	delete[] bruhat_ascents_by_descents;
}



/*
 * Function: CheckAutomorphism_Helper
 * ---------------------------------------------------------
 * Function returns true if st has same order as s't' in Coxeter
 * group W, where s' and t' and the generators indexed by the
 * assignments s' = (word[s] - 1) and t' = (word[t] - 1). 
 *
 */
bool CheckAutomorphism_Helper(interactive::CoxGroup* W, const CoxWord &word, Generator s, Generator t)
{
	CoxNbr identity = 0;
	
	CoxNbr x = identity;
	CoxNbr y = identity;
	
	while(true)
	{
		W->lprod(x,s);
		W->lprod(x,t);
		
		W->lprod(y,word[s]-1);
		W->lprod(y,word[t]-1);
		
		
		if(x == identity)
		{
			if(y == identity) break;
			else return false;
		}
		
		if(x != identity && y == identity) return false;
	}
	
	return true;
}



/*
 * Function: CheckAutomorphism
 * ---------------------------------------------------------
 * Function checks whether an input CoxWord can be interpreted
 * as an involution of the Coxeter group W. This happens if the
 * word is a permutation of 123...r where r is the rank of W, and
 * this permutation has order one or two, and if for all i and j
 * the product of the permuted generators in positions i and j 
 * has the same order as the product of generators i and j.
 *
 * Note that the permutations of 123...r are all distinct reduced 
 * expressions and so any automorphism of W can be specified by a 
 * single CoxWord.
 *
 */
bool CheckAutomorphism(interactive::CoxGroup* W, CoxWord &word)
{
	if(ERRNO) return false;
	
	int r = W->rank();
	
	// Reinterprets user input "" as identity automorphism 1234...n
	if(word.length() == 0)
	{
		for(Generator s=0; s<r; s++) word.append(s+1);
		return true;
	}
	
	// If input word does not have correct length, cannot represent automorphism
	if(word.length() != r) return false;
	
	
	// Following code checks that every generator appears in input word exactly once
	std::set<CoxLetter> distinct;
	for(int i=0; i<word.length(); i++) 
	{
		// Here checks that input word represents involution of generating set
		if(i+1 != word[word[i]-1]) return false;
		
		distinct.insert(word[i]);
	}
	if(distinct.size() != r) return false;
	
	
	// At this point, have checked that input word representats permutation of
	// simple generating set, with order one or two. Finally, check that this 
	// permutation induces automorphism of Coxeter group	
	for(Generator s=0; s<r; s++)
	{
		for(Generator t=0; t<r; t++)
		{
			if(! CheckAutomorphism_Helper(W,word,s,t)) return false;
		}
	}
	
	return true;
}



/*
 * Function: GetGroupAutomorphism
 * ---------------------------------------------------------
 * Queries the user for a permutation of 123...r, where r is
 * the rank of our Coxeter group W, which will specifies an 
 * involution of W.
 *
 * The identity automorphism is used if the user enters as
 * input the empty string.
 *
 */
void TwistedInvolutionsPoset::GetGroupAutomorphism()
{	
	cout << endl;
	while(true)
	{
		cout << "involution : ";
		CoxWord word = interactive::getCoxWord(W);
		
		if(CheckAutomorphism(W,word))
		{
			cout << endl << "(Setting involution to ";
			
			star = CoxWord(0);
			for(int i=0; i<word.length(); i++) 
			{
				cout << int(word[i]);
				star.append(word[i]-1);
			}
		
			cout << ")" << endl;			
			break;
		}
		
	}

	// Sets the group name prefix for all output file names 
	SetGroupName();
}



/*
 * Function: SetGroupName
 * ---------------------------------------------------------
 * Helper function which sets the group name prefix for all 
 * output file names. This prefix generally has the form
 *
 *   (Type)(Rank)_(Involution)
 *
 * For example, "A5_12345" or "F4_4321" or "I2(7)_12"
 *
 */
void TwistedInvolutionsPoset::SetGroupName()
{
	string type(W->type().name().ptr());
	
	stringstream ss;
	ss << type;
	ss << W->rank();
	
	// Special case handing for dihedral type
	if(type == "I") ss << "(" << (W->order()/2) << ")";
	
	ss << "_";
	for(int i=0; i<star.length(); i++) ss << int(star[i]+1);
	
	group_name = ss.str();
}



/*
 * Function: GetInvolutions
 * ---------------------------------------------------------
 * Adds all CoxNbr's corresponding to twisted involutions to 
 * a vector maintained by the TwistedInvolutionsPoset. A map 
 * from CoxNbr's to list indices is updated for use in some 
 * other methods called by TwistedInvolutionsPoset constructor.
 *
 */
void TwistedInvolutionsPoset::GetInvolutions(std::map<CoxNbr,int> &index_map)
{
	for (CoxNbr y = 0; y < W->order(); ++y) 
	{
		if (GetStarInverse(y) == y) 
		{
			index_map[y] = involutions.size();
			involutions.push_back(y);
		}
	}
}



/*
 * Function: ComputeActions
 * ---------------------------------------------------------
 * Helper function creates 2D array elem_actions whose (i,s) entry 
 * is the index of the twisted involution w', where i is the 
 * index of the twisted involution w and s is a simple generator and
 *
 *   w' = sws* or sw, whichever is a twisted involution distinct from w.
 *
 * Also creates 1D array elem_lengths whose ith entry is the length of w.
 *
 */
void TwistedInvolutionsPoset::ComputeActions(const std::map<CoxNbr,int> &index_map)
{	
	elem_actions = new int*[size()];
	elem_lengths = new int[size()];
	
	for(int i = 0; i < size(); i++)
	{
		CoxNbr w = involutions[i];
		elem_lengths[i] = W->length(w);
		
		elem_actions[i] = new int[W->rank()];
		for(Generator s = 0; s < W->rank(); s++)
		{
			CoxNbr ws = w;
			CoxNbr sw = w;
			CoxNbr s_act_w;
			
			W->lprod(sw,s);
			W->prod(ws,star[s]);
			
			s_act_w = sw;
			if(sw != ws) W->prod(s_act_w,star[s]);
			
			elem_actions[i][s] = index_map.find(s_act_w)->second;
		}
	}
	
}



/*
 * Function: RecordIndexing
 * ---------------------------------------------------------
 * Creates two text files in the INVOLUTION_INDICES_FOLDER
 * and ALL_ELEMENT_INDICES_FOLDER. The nth line of the first 
 * (respectively, second)  file contains a reduced expression 
 * for the nth element in our set of twisted involutions I_* 
 * (respectiely, the nth element in our Coxeter group W). 
 *
 */
void TwistedInvolutionsPoset::RecordIndexing() const
{
	RecordIndexing_Helper(Get_InvolutionIndices_FileName(), involutions);
	
	std::vector<CoxNbr> all;
	for(CoxNbr w = 0; w < W->order(); w++) all.push_back(w);
	
	RecordIndexing_Helper(Get_AllElementIndices_FileName(), all);	
}



/*
 * Function: RecordIndexing_Helper
 * ---------------------------------------------------------
 * Helper function called by RecordIndexing; prints list of 
 * reduced expressions in order to a file.
 *
 */
void TwistedInvolutionsPoset::RecordIndexing_Helper(const string& filename, const std::vector<CoxNbr> &list) const
{
	FILE *output = fopen(filename.c_str(), "w");
	
	for (int i = 0; i < list.size(); i++) 
	{
		W->print(output,list[i]);
		fprintf(output,"\n");
	}
	
	fclose(output);	
}



/*
 * Function: SetupBruhatOrderData
 * ---------------------------------------------------------
 * Initiates several helper methods to collect various data 
 * associated with the Bruhat order on W and I_*.
 *
 */
void TwistedInvolutionsPoset::SetupBruhatOrderData()
{
	ComputeBruhatOrder();
	ComputeBruhatAscents();
	ComputeBruhatAscentsByDescents();
	ComputeBruhatIntervals_in_W();
}



/*
 * Function: ComputeBruhatOrder
 * ---------------------------------------------------------
 * This helper function, called by the constructor, computes 
 * the boolean array bruhat_order, which contains "true" in 
 * entry (i,j) if y := involutions[i] and w := involutions[j] 
 * have y <= w in the Bruhat order of W.
 *
 */
void TwistedInvolutionsPoset::ComputeBruhatOrder()
{
	bruhat_order = new bool*[size()];
	for(int y_index = 0; y_index < size(); y_index++)
	{
		bruhat_order[y_index] = new bool[size()];
		for(int w_index = 0; w_index < size(); w_index++)
		{
			bruhat_order[y_index][w_index] = W->inOrder(involutions[y_index], involutions[w_index]);
			
			// Makes sure that the total order on involutions induced by their indices
			// extends the Bruhat order, and aborts of this is not the case
			if(bruhat_order[y_index][w_index] && w_index < y_index) 
			{
				SafePolynomial::AnnounceErrorMessage("Indexing does not induce total order compatible with Bruhat order");
			}
		}
	}
}



/*
 * Function: ComputeBruhatAscents
 * ---------------------------------------------------------
 * This helper method, called by the constructor, sets up the 
 * private variable bruhat_ascents. This is an array of vectors
 * whose ith entry is the list of integers which are the indices 
 * of the twisted involutions which cover involutions[i] in the 
 * subposet (I_*, <) induced the Bruhat order on W.
 *
 */
void TwistedInvolutionsPoset::ComputeBruhatAscents()
{
	bruhat_ascents = new std::vector<int>[size()];
	
	
	// Sets up map with key value pairs (i,vec), where
	// i is an integer and vec is the list of indices 
	// of the twisted involutions with rank i in (I_*, <)
	//
	std::map<int,std::vector<int> > involutions_by_rank;
	for(int y=0; y<size(); y++) involutions_by_rank[GetInvolutiveRank(y)].push_back(y);
	
	
	int rho = 0;
	int max_rho = involutions_by_rank.size() - 1;
	while(rho < max_rho)
	{
		ComputeBruhatAscents_Helper(involutions_by_rank,rho);
	}
}



/*
 * Function: ComputeBruhatAscents_Helper
 * ---------------------------------------------------------
 * This helper method computes the covers in the poset (I_*, <)
 * of all twisted involutions of a particular fixed rank. 
 *
 */
void TwistedInvolutionsPoset::ComputeBruhatAscents_Helper(const std::map<int,std::vector<int> > &involutions_by_rank, int &rho)
{
	std::vector<int> curr = involutions_by_rank.find(rho)->second;
	std::vector<int> next = involutions_by_rank.find(rho+1)->second;
	
	for(int i=0; i<curr.size(); i++) 
	{
		for(int j=0; j<next.size(); j++) 
		{
			int y_index = curr[i];
			int z_index = next[j];
			
			if(BruhatLessOrEqual(y_index,z_index)) bruhat_ascents[y_index].push_back(z_index);
		}
	}
	
	rho++;
}



/*
 * Function: GetInvolutiveRank
 * ---------------------------------------------------------
 * Helper function returns the rank of the twisted involution 
 * indexed by w_index in the poset I_* of twisted involutions 
 * with respect to the Bruhat order.
 *
 */
int TwistedInvolutionsPoset::GetInvolutiveRank(const int &w_index) const
{
	Generator s;
	if(! GetLeftDescent(w_index,s)) return 0;
	
	return 1 + GetInvolutiveRank(ActOnInvolution(w_index,s));
}



/*
 * Function: ComputeBruhatAscentsByDescents
 * ---------------------------------------------------------
 * Helper function sets up the private field bruhat_ascents_by_descents,
 * a 3D-array whose (s,w,y) entry is the set of indices of twisted
 * involutions z with y <= z <= w and sw < w.
 *
 */
void TwistedInvolutionsPoset::ComputeBruhatAscentsByDescents()
{
	bruhat_ascents_by_descents = new std::vector<int>**[W->rank()];
	
	for(Generator s=0; s<W->rank(); s++)
	{
		bruhat_ascents_by_descents[s] = new std::vector<int>*[size()];
		for(int w = 0; w < size(); w++)
		{	
			bruhat_ascents_by_descents[s][w] = new std::vector<int>[w+1];
			for(int y = 0; y <= w; y++)
			{
				// Fills in bruhat_ascents_by_descents[s][w][y] 
				ComputeBruhatAscentsByDescents_Helper(s,w,y);
			}
		}
	}
}



/*
 * Function: ComputeBruhatAscentsByDescents_Helper
 * ---------------------------------------------------------
 * Helper function called by ComputeBruhatAscentsByDescents. 
 * This method explicitly computes the set of indices of
 * twisted involutions z with y <= z <= w and sw < w.
 *
 */
void TwistedInvolutionsPoset::ComputeBruhatAscentsByDescents_Helper(Generator s, int w, int y)
{
	if(IsLeftDescent(y,s) && BruhatLessOrEqual(y,w))
	{
		std::vector<int> vec = bruhat_ascents[y];
		for(int i=0; i<vec.size(); i++)
		{
			int z = vec[i];
			if(IsLeftDescent(z,s) && BruhatLessOrEqual(z,w))
			{
				bruhat_ascents_by_descents[s][w][y].push_back(z);
			}
		}
	}
}



/*
 * Function: ComputeBruhatIntervals_in_W
 * ---------------------------------------------------------
 * Sets up the lower_intervals_data struct stored by object.
 *
 */
void TwistedInvolutionsPoset::ComputeBruhatIntervals_in_W()
{
	fcoxgroup::FiniteCoxGroup* fg = dynamic_cast<fcoxgroup::FiniteCoxGroup*>(W);
	fg->fullContext(); 
	fg->fillMu();
	
	wgraph::WGraph X(0);
	cells::lWGraph(X,fg->kl());
	
	lower_intervals_data.AllocateLists(X.size(), fg->rank());
	lower_intervals_data.FillLists(X);
}



/*
 * Function: GetStarInverse
 * ---------------------------------------------------------
 * Given a CoxNbr representing an element x of W, returns the
 * CoxNbr representing the element (x*)^-1. The input x is not 
 * passed by reference as it is modified within the function.
 *
 */
CoxNbr TwistedInvolutionsPoset::GetStarInverse(CoxNbr x) const
{ 
	CoxNbr ans = 0;
	Generator s;
	
	while(W->length(x) > 0) 
	{
		Generator s = s = firstBit(W->ldescent(x));
		W->lprod(x,s);
		W->lprod(ans,star[s]);
	}
	
	return ans;
}



/*
 * Function: GetLeftDescent
 * ---------------------------------------------------------
 * For w = involutions[w_index], returns false if w = 1 and 
 * true otherwise. In the latter case the reference s is 
 * assigned to a left descent of w.
 *
 */
bool TwistedInvolutionsPoset::GetLeftDescent(const int &w_index, Generator &s) const
{
	if(GetLength(w_index) == 0) return false;
	
	s = firstBit(W->ldescent(involutions[w_index]));
	
	return true;
}



/*
 * Function: IsLeftDescent
 * ---------------------------------------------------------
 * Returns true if input generator s is a left descent of the twisted
 * involution indexed by w_index in our TwistedInvolutionsPoset
 *
 */
bool TwistedInvolutionsPoset::IsLeftDescent(const int &w_index, const Generator &s) const 
{ 
	return GetLength(ActOnInvolution(w_index,s)) < GetLength(w_index); 
}



/*
 * Function: Commutes
 * ---------------------------------------------------------
 * Returns true if sw=ws, false otherwise, where w is the twisted
 * involution with index w_index in our TwistedInvolutionsPoset.
 * Here as usual, we define w := involutions[w_index].
 *
 */
bool TwistedInvolutionsPoset::Commutes(const int &w_index, const Generator &s) const 
{ 
	return (GetLength(ActOnInvolution(w_index,s)) - GetLength(w_index)) % 2 != 0; 
} 




/*
 * Function: InitializeInterval
 * ---------------------------------------------------------
 * Return an InvoBruhatIterator object whcih will iterate over all
 * twisted involutions z with y <= z <= w and s in Des_L(w).
 *
 */
InvolBruhatIterator TwistedInvolutionsPoset::InitializeInterval(const int &y, const int &w, const Generator &s) const
{
	if( !IsLeftDescent(y,s) ) 
	{
		SafePolynomial::AnnounceErrorMessage("Bruhat Interval Iterator Error: minimal element y must have s as left descent");
	}
	
	// In this case returns iterator over empty interval
	if( !BruhatLessOrEqual(y,w) ) 
	{
		return InvolBruhatIterator();
	}
	
	return InvolBruhatIterator(bruhat_ascents_by_descents[s], y, w);
}



/*
 * Function: PrintPercentProgress
 * ---------------------------------------------------------
 * Helper method increments a counter variable and prints to 
 * the command line the percentage progress whenever this %
 * is divisible by an input increment.
 *
 */
void TwistedInvolutionsPoset::PrintPercentProgress(int &done, int total, int incr) const
{
	int percent = total/100 == 0 ? 1 : total/100;
	if(++done % (incr*percent) == 0) 
	{
		if(verbose_printing_on) printf("  %d%%\n",(done*100)/total);
	}
	
	if(done == total && verbose_printing_on) cout << endl;
}



/*
 * Function: PrintAllProgress
 * ---------------------------------------------------------
 * Helper method increments a counter variable and prints on 
 * each call a progress update to the command line output.
 * Functions use this method during slow computation when it
 * seeing every increment of progress more useful than periodic
 * percentage updates.
 *
 */
void TwistedInvolutionsPoset::PrintAllProgress(const string& space, int i, int total) const
{
	if(verbose_printing_on) 
	{
		cout << space << (i+1) << " of " << total << endl;
		if(i+1 == total) cout << endl;
	}
}







/*
 * Function: Sigma_Coeff_Helper
 * ---------------------------------------------------------
 * Helper function returns coefficient of the monomial
 *
 *   q^{( ell(w) - ell(y) - offset ) / 2}
 * 
 * in polynomial Psigma_{y,w} stored in array psigmas.
 *
 */
SafeCoeff TwistedInvolutionsPoset::Sigma_Coeff_Helper(SafePolynomial** psigmas, const int &y, const int &w, int offset) const
{
	int diff = GetLength(w) - GetLength(y) - offset;
	if(diff % 2 != 0) return 0;
	
	return psigmas[y][w][diff/2];	
}



/*
 * Function: Mu_Sigma_Coeff
 * ---------------------------------------------------------
 * This function computes the integer mu^sigma(y,w;s) as defined 
 * in the accompanying paper. The return value is a SafePolynomial
 * since we have implemented this type to detect any integer 
 * overflow errors which occur (and abort the program in this case).
 *
 */
SafePolynomial TwistedInvolutionsPoset::Mu_Sigma_Coeff(SafePolynomial** psigmas, const int &y, const int &w, const Generator &s) const
{
	SafePolynomial mu = SafePolynomial::GetZero();
	
	mu += Nu_Sigma_Coeff(psigmas, y,w);	
	if(Commutes(y,s)) mu += Mu_Sigma_Coeff(psigmas, ActOnInvolution(y,s), w);
	if(Commutes(w,s)) mu -= Mu_Sigma_Coeff(psigmas, y, ActOnInvolution(w,s));
		
	
	InvolBruhatIterator iterator = InitializeInterval(y,w,s);
	while(iterator.hasNext())
	{
		CoxNbr x = iterator.next();
		if(x == y || x == w) continue;
		
		mu -= (Mu_Sigma_Coeff(psigmas,y,x) * Mu_Sigma_Coeff(psigmas,x,w));
	}
	
	return mu;
}



/*
 * Function: ComputeMsigmas_Helper
 * ---------------------------------------------------------
 * Helper function called by ComputeMsigmas.
 *
 */
SafePolynomial TwistedInvolutionsPoset::ComputeMsigmas_Helper(SafePolynomial** psigmas, const int &y, const int &w, const Generator &s) const
{
	if((GetLength(w) - GetLength(y)) % 2 != 0) 
	{
		return SafePolynomial(1, Mu_Sigma_Coeff(psigmas,y,w));
	}
	else return Mu_Sigma_Coeff(psigmas,y,w,s);
}



/*
 * Function: ComputeMsigmas
 * ---------------------------------------------------------
 * Returns a 3D-array of SafePolynomials msigmas, which stores
 * as msigmas[y][w][s] the Laurent polynomial 
 *
 *   m^{\sigma}( y \xrightarrow{s} w )
 *
 * defined in the multiplication formula for C_s * A_w.
 * 
 * Two important but somewhat technical caveats:
 *
 *   1. This function is only called after the polynomial Psigma_{y,w} 
 *		have been computed and printed to output. 
 *
 *   2. The Laurent polynomials m^sigma( y -s-> w ) have the form
 *		c or c*(v + v^{-1}) where c is an integer. What is stored
 *		in the array msigmas is only the terms of nonnegative degree,
 *		i.e., polynomials of the form c or c*x.
 *
 */
SafePolynomial*** TwistedInvolutionsPoset::ComputeMsigmas() const
{
	SafePolynomial** psigmas = SafePolynomial::ReadSafePolynomialArray( Get_Psigma_Polynomials_FileName(), size() );
	
	SafePolynomial*** msigmas = new SafePolynomial**[size()];
	for(int y=0; y<size(); y++)
	{
		msigmas[y] = new SafePolynomial*[size()];
		
		for(int w=0; w<size(); w++)
		{
			msigmas[y][w] = new SafePolynomial[W->rank()];
			
			for(Generator s = 0; s<W->rank(); s++)
			{
				if(IsLeftDescent(y,s)) msigmas[y][w][s] = ComputeMsigmas_Helper(psigmas,y,w,s);
				else msigmas[y][w][s] = SafePolynomial::GetZero();
			}
		}
	}	
	
	SafePolynomial::Delete_2D_PolynomialArray(psigmas,size());
	return msigmas;
}







/*
 * Function: Compute_Psigma_Polynomials
 * ---------------------------------------------------------
 * Computes the twisted Kazhdan-Lusztig polynomials Psigma_{y,w} 
 * for each pair of twisted involutions y,w in I. Output is 
 * printed to a text file in the folder labeled by 
 * Psigma_POLYNOMIALS_LABEL in the output directory. Each line 
 * of the output file has the form
 *
 *   i,j;a,b,c,d...
 *
 * to indicate that if y and w are the ith and jth elements of
 * the set of twisted involutions I, then 
 *
 *   P_{y,w} = a + bq + cq^2 + dq^3 + ...
 *
 */
void TwistedInvolutionsPoset::Compute_Psigma_Polynomials() const
{
	cout << "Computing polynomials Psigma_{y,w}." << endl;
	
	// Counter variables for command line progress updates
	int total = involutions.size()*(involutions.size()+1)/2, done = 0;
	
	
	SafePolynomial** psigmas = SafePolynomial::InitializeSafePolynomialArray(size());
	
	
	// The recurrence we have to compute Psigma_{y,w} requires that
	// Psigma_{y',w'} be computed already if either 
	//
	//   w' < w		or		w = w' and y < y'
	//
	// Hence the iteration order here.
	//
	for(int w=0; w<involutions.size(); w++) 
	{
		for(int y=w; y>=0; y--) 
		{
			if(BruhatLessOrEqual(y,w)) 
			{
				Compute_PsigmaPolynomial_Helper(psigmas,y,w);
			}
			
			PrintPercentProgress(done,total);
		}
	}
	
	
	// Prints computation to output file and frees memory 
	SafePolynomial::WriteSafePolynomialArray(Get_Psigma_Polynomials_OutputFile(), size(), psigmas);	
	SafePolynomial::Delete_2D_PolynomialArray(psigmas,size());
}



/*
 * Function: Extract_Psigma
 * ---------------------------------------------------------
 * This function is called in Compute_PsigmaPolynomial_Helper
 * after the recurrence ... has been calculated.
 *
 * At this point we have two twisted involutions y and w and
 * a generator s, we wish to extract the true value of the
 * polynomial Psigma_{y,w} from the input "next".
 *
 * In the situation that sw=ws*, the polynomial stored as 
 * "next" has the form
 *
 *   a_0 + (a_0 + a_1)q + (a_1 + a_2)q^2 + ... + (a_{n-1} + a_n)q^n + higher order terms
 *
 * where n = floor( (ell(w) - ell(y) - 1) / 2 ) and where the 
 * coefficients a_i satisfy
 * 
 *   Psigma_{y,w} = a_0 + a_1q + a_2q^2 + ... + a_nq^n.
 *
 * The following method extracts this polynomial from the input 
 * and stores it in the variable "next".
 *
 */
void Extract_Psigma(bool w_s_commutes, int length_difference, SafePolynomial &next)
{
	// If sw != ws then next = Psigma_{y,w}, so return
	if(!w_s_commutes) return;
	
	// Computes highest possible degree of Psigma_{y,w}
	int e = (length_difference-1)/2;
	
	// Modifies next to have form a_0 + a_1q + ... + a_nq + higher order terms
	SafeCoeff coeff = 0;
	for(int i=0; i<=e; i++) 
	{
		next -= SafePolynomial(i,coeff);
		coeff = next[i];
	}
	
	// Removes extraneous higher order terms from next
	int deg = next.Degree();
	for(int i=e+1; i<=deg; i++) 
	{
		next -= SafePolynomial(i,next[i]);
	}
}



/*
 * Function: Compute_PsigmaPolynomial_Helper
 * ---------------------------------------------------------
 * Implements the recurrence which determines Psigma_{y,w}, 
 * provided we have computed Psigma_{y',w'} for all (y',w')
 * such that w' < w or such that w' = w and y < y'.
 *
 */
void TwistedInvolutionsPoset::Compute_PsigmaPolynomial_Helper(SafePolynomial** psigmas, const CoxNbr &y, const CoxNbr &w) const
{	
	// If y == w then Psigma_{y,w} = 1
	if(y == w)  
	{
		psigmas[y][w] = SafePolynomial(0);
		return;
	}
	
	// Choose a descent of w and computes s ltimes w, s ltimes y
	Generator s;
	GetLeftDescent(w,s);
	CoxNbr sy = ActOnInvolution(y,s);
	CoxNbr sw = ActOnInvolution(w,s);
	
	// If s is not also a descent of y then Psigma_{y,w} = Psigma_{s ltimes y,w}
	if(!IsLeftDescent(y,s)) 
	{
		psigmas[y][w] = psigmas[sy][w];
		return;
	}
	
	// Otherwise we compute Psigma_{y,w} using the recurrence 
	// given in the accompanying paper
	//
	AddLeadingTerms(psigmas[sy][sw], psigmas[y][sw], Commutes(y,s), psigmas[y][w]);
	
	InvolBruhatIterator iterator = InitializeInterval(y,w,s);
	while(iterator.hasNext())
	{
		CoxNbr z = iterator.next();
		if(z == w) continue;
		
		Subtract_Sum_z_Terms(psigmas, s, z, w, psigmas[y][z], psigmas[y][w]);
	}
	
	Extract_Psigma(Commutes(w,s), GetLength(w) - GetLength(y), psigmas[y][w]);
}



/*
 * Function: AddLeadingTerms
 * ---------------------------------------------------------
 * Helper function whcih adds to the currently uninitialized 
 * value of Psigma_{y,w} the leading terms 
 *
 *   a(q)*Psigma_{y',w'} + b(q)*Psigma_{y,w'}
 *
 * where y' = s ltimes w and w' = s ltimes w and 
 *
 *   a(q) = q+1		if sy=ys* and 1	  otherwise
 *   b(q) = q^2+q	if sy=ys* and q^2 otherwise 
 *
 */
void TwistedInvolutionsPoset::AddLeadingTerms(const SafePolynomial& Psigma_sy_sw, const SafePolynomial& Psigma_y_sw, bool y_s_commutes, SafePolynomial &next) const
{	
	SafePolynomial first, second;
	
	if(y_s_commutes) 
	{
		first = SafePolynomial(1);
		first += SafePolynomial(0);
		
		second = SafePolynomial(2);
		second -= SafePolynomial(1);
	}
	else
	{
		first = SafePolynomial(0);
		second = SafePolynomial(2);
	}	
	
	first *= Psigma_sy_sw;
	second *= Psigma_y_sw;
	
	next += first;
	next += second;
}



/*
 * Function: Subtract_Sum_z_Terms
 * ---------------------------------------------------------
 * Helper function whcih subtracts the currently stored value 
 * of Psigma_{y,w} the summation
 *
 *   sum_{z} v^{ell(w) - ell(z) + c} * m^sigma( z -s-> w') * Psigma_{y,z}
 *
 * where the sum is over all z in I_* with sz<z (and y <= z < w, 
 * though if this last condition is not satisfied then the terms
 * are automatically zero). Here w' = s ltimes w.
 *
 */
void TwistedInvolutionsPoset::Subtract_Sum_z_Terms(SafePolynomial** psigmas, const Generator &s, const int &z, const int &w, const SafePolynomial & Psigma_y_z, SafePolynomial &next) const
{
	int len_diff = GetLength(w) - GetLength(z) + (Commutes(w,s) ? 1 : 0);
	CoxNbr sw = ActOnInvolution(w,s);
	
	SafePolynomial other_term;
	if( len_diff % 2 == 0 ) 
	{
		other_term = SafePolynomial( len_diff / 2 );
		other_term *= Mu_Sigma_Coeff(psigmas, z, sw, s);
	} 
	else 
	{
		other_term = SafePolynomial( (len_diff - 1) / 2 );
		other_term += SafePolynomial( (len_diff + 1) / 2 );
		other_term *= Mu_Sigma_Coeff(psigmas, z, sw);	
	}
	
	other_term *= Psigma_y_z;
	next -= other_term;
}








/*
 * Function: Compute_P_Polynomials
 * ---------------------------------------------------------
 * Computes the Kazhdan-Lusztig polynomials P_{y,w} for each 
 * pair of twisted involutions y,w in I. Output is printed to 
 * a text file in the folder labeled by P_POLYNOMIALS_LABEL 
 * in the output directory. The output is formatted exactly 
 * as in the Compute_Psigma_Polynomials function.
 *
 */
void TwistedInvolutionsPoset::Compute_P_Polynomials() const
{
	cout << "Computing polynomials P_{y,w}." << endl;
	
	
	// Counter variables for command line progress updates
	int total = size() * size(), done = 0;
	
	
	FILE *output = Get_P_Polynomials_OutputFile();
	
	for(int i=0; i<size(); i++) {
		for(int j=0; j<size(); j++) {
			
			CoxNbr y = involutions[i];
			CoxNbr w = involutions[j];
			
			if(BruhatLessOrEqual(i,j)) 
			{
				PrintPolynomial(output, i, j, W->klPol(y,w));
			}
			
			PrintPercentProgress(done,total);
		}
	}
	
	fclose(output);
}








/*
 * Function: Compute_hsigma_StructureConstants
 * ---------------------------------------------------------
 * Computes the Laurent polynomials h^sigma_{x,y;z} for each 
 * x in W and twisted involutions y,z in W. These polynomials
 * are the structure constants such that
 *
 *   C_x A_y = sum_z h^sigma_{x,y;z} A_z
 * 
 * The polynomials are printed to a series of text files in 
 * the folder labeled by hsigma_STRUCTURECONSTANTS_LABEL in 
 * the output directory. Each output text file name ends in 
 * a nonnegative integer; for example, if W = H3 and * = 123 
 * then one output file would be named 
 *
 *   H3_123_hsigma_StructureConstants_26
 *
 * Each line of this output file has the form
 *
 *   i,j;a,b,c,d...
 *
 * to indicate that if y is the 26th element of I and
 * x is the ith element of W and z is the jth element of I
 * then
 *
 *   h^sigma_{x,y;z} = a + b(v + v^{-1}) + c(v^2 + v^{-2}) + d(v^3 + v^{-3}) + ...
 *
 */
void TwistedInvolutionsPoset::Compute_hsigma_StructureConstants() const
{
	cout << "Computing polynomials hsigma_{x,y;z}." << endl;
	
	
	// First computes the polynomials h^sigma_{s,y;z} with s = a simple generator 
	InvolIndexPol_Map** C_s_actions = Get_C_s_actions();
	
	
	// Iterates over twisted involutions and computes the family h^sigma_{x,y;z}
	// with y fixed. Output is printed to a separate text file for each y
	for(int y=0; y<size(); y++) 
	{
		PrintAllProgress("", y, size());
		Compute_hsigma_StructureConstants_Helper(C_s_actions, y);
	}
	
	// Deletes the array C_s_actions
	Delete_2D_VectorArray(C_s_actions, W->rank());
}



/*
 * Function: Get_C_s_actions
 * ---------------------------------------------------------
 * Returns a 2D-array of InvolIndexPol_Maps whose (s,w) entry represents
 * the vector in M_{q^2} given by the product C_s * A_w 
 *
 */
InvolIndexPol_Map** TwistedInvolutionsPoset::Get_C_s_actions() const
{
	// y, w, s
	SafePolynomial*** msigmas = ComputeMsigmas();
	InvolIndexPol_Map** C_s_actions = new InvolIndexPol_Map*[W->rank()];
	
	for(Generator s = 0; s < W->rank(); s++)
	{
		C_s_actions[s] = new InvolIndexPol_Map[size()];		
		for(int w = 0; w < size(); w++) 
		{
			Get_C_s_actions_Helper(C_s_actions[s][w], s, w, msigmas);	
		}
	}
	
	Delete_3D_PolynomialArray(msigmas, size());
	return C_s_actions;
}



/*
 * Function: IncrementPolynomialMap
 * ---------------------------------------------------------
 * Add the input polynomial pol to the entry with key w in 
 * the map output. If this map does not contain any key w
 * then the function inserts pol with key w into the output.
 *
 */
void IncrementPolynomialMap(InvolIndexPol_Map &output, const int &w, SafePolynomial pol)
{
	if(output.find(w) == output.end()) output[w] = SafePolynomial::GetZero();
	output[w] += pol;
}



/*
 * Function: Get_C_s_actions_Helper
 * ---------------------------------------------------------
 * Helper function called by Get_C_s_actions which assigns the
 * parameter actions to the InvolIndexPol_Map which represents 
 * the vector in M_{q^2} given by C_s A_w. 
 *
 */
void TwistedInvolutionsPoset::Get_C_s_actions_Helper(InvolIndexPol_Map &actions, const Generator &s, const int &w, SafePolynomial*** msigmas) const
{
	actions = InvolIndexPol_Map();
	
	if(IsLeftDescent(w,s))
	{
		// In this case C_s A_w = (q + q^{-1}) A_w
		//
		IncrementPolynomialMap(actions, w,  SafePolynomial(2)); 
	}
	
	else 
	{	
		// Otherwise either C_s A_w = (v + v^{-1}) A_{sw} + (sum ...) if sw=ws*
		//				 or C_s A_w = A_{sws*} + (sum ...) otherwise
		//
		CoxNbr sw = ActOnInvolution(w,s);
		SafePolynomial leading_coeff = Commutes(w,s) ? SafePolynomial(1) : SafePolynomial(0);
		
		
		IncrementPolynomialMap(actions, sw, leading_coeff);
		
		// The following loops adds sum_{y in I_*, sy<y<s ltimes w} m^sigma( y -s-> w) A_y
		//
		InvolBruhatIterator iterator = InitializeInterval(ActOnInvolution(0,s), sw, s);
		while(iterator.hasNext())
		{
			CoxNbr y = iterator.next();
			if(y != sw) 
			{
				IncrementPolynomialMap(actions, y,  msigmas[y][w][s]);
			}
		}
	}
}	



/*
 * Function: Compute_hsigma_StructureConstants_Helper
 * ---------------------------------------------------------
 * Helper function called by Compute_hsigma_StructureConstants which computes
 * the Laurent polynomials h^sigma_{x,y;z} for a fixed twisted involution y.
 *
 */
void TwistedInvolutionsPoset::Compute_hsigma_StructureConstants_Helper(InvolIndexPol_Map** C_s_actions, int y) const
{
	// Counter variables for command line progress updates
	int total = W->order(), done = 0;
	
	
	InvolIndexPol_Map* hsigmas = new InvolIndexPol_Map[W->order()];	
	for(CoxNbr x = 0; x < W->order(); x++)
	{
		PrintPercentProgress(done,total);
		
		// Base case x = 1
		hsigmas[x] = InvolIndexPol_Map();
		if(x == 0) 
		{
			hsigmas[x][y] = SafePolynomial(0);
		}
		else
		{
			Generator s = firstBit(W->ldescent(x));
			CoxNbr sx = x;
			W->lprod(sx,s);
			
			// We compute C_x A_y using the inductive formula
			//
			//   C_x A_y = C_s (C_{sx} A_y) - sum_{sz < z < sx} mu(z,sx) C_z A_y
			//
			// where s is a descent of x. Below, the first line 
			// computes C_s (C_{sx} A_y) and the second line subtracts 
			// the terms sum_{sz < z < w} mu(z,w) C_z A_y from the result
			//
			Multiply_by_C_s(hsigmas[x], hsigmas[sx], C_s_actions[s]);
			Subtract_Extra_C_z_Terms(hsigmas[x], hsigmas, x, sx, s);
		}
	}
	
	// Prints results to output file
	Write_hsigma_Array(y, hsigmas); 
	delete[] hsigmas;
}



/*
 * Function: Combine_HalfPolynomial_Maps
 * ---------------------------------------------------------
 * To explain this function, some notation: given polynomials
 *
 *   a = a_0 + a_1x + a_2x^2 + ...
 *   b = b_0 + b_1x + b_2x^3 + ...
 *
 * define
 *
 *   a' = a_0 + a_1(x + 1/x) + a_2(x^2 + 1/x^2) + ...
 *   b' = b_0 + b_1(x + 1/x) + a_2(x^2 + 1/x^2) + ...
 *
 * and let m(a,b) be the unique polynomial in Z[x] such that 
 * a'*b' - m(a,b) lies in (1/x)Z[1/x]. This function computes 
 * 
 *   "current += m(pol,summand)"
 *
 * where we interprent the maps current and summand as vectors 
 * of polynomials and carry out addition and m componentwise.
 *
 */
void Combine_HalfPolynomial_Maps(InvolIndexPol_Map &current, const InvolIndexPol_Map &summand, const SafePolynomial &pol)
{
	InvolIndexPol_Map::const_iterator it;
	for(it = summand.begin(); it != summand.end(); ++it) 
	{
		SafePolynomial next = SafePolynomial::MultiplyHalfPolynomials(it->second, pol);
		
		IncrementPolynomialMap(current, it->first, next);
	}	
}

void Combine_HalfPolynomial_Maps(InvolIndexPol_Map &current, const InvolIndexPol_Map &summand, int constant)
{
	Combine_HalfPolynomial_Maps(current, summand, SafePolynomial(0,constant));
}



/*
 * Function: Multiply_by_C_s
 * ---------------------------------------------------------
 * Multiplies the InvolIndexPol_Map prev (representing a vector in M_{q^2})
 * by C_s in H_{q^2} for a fixed value of s (which is determined by the 
 * value of C_s_actions passed as an argument), then adds this vector to
 * the InvolIndexPol_Map next.
 *
 */
void TwistedInvolutionsPoset::Multiply_by_C_s(InvolIndexPol_Map &next, const InvolIndexPol_Map &prev, InvolIndexPol_Map* C_s_actions) const
{
	InvolIndexPol_Map::const_iterator it;
	for(it = prev.begin(); it != prev.end(); ++it)
	{
		CoxNbr w = it->first;
		SafePolynomial coeff = it->second;
		
		Combine_HalfPolynomial_Maps(next, C_s_actions[w], coeff);
	}
}



/*
 * Function: Subtract_Extra_C_z_Terms
 * ---------------------------------------------------------
 * Subtracts from the vector in M_{q^2} represented by next the sum
 *
 *   sum_{sz < z < sx} mu(z,sx) C_z A_y
 *
 * where s is a descent of x in W. 
 *
 */
void TwistedInvolutionsPoset::Subtract_Extra_C_z_Terms(InvolIndexPol_Map &next, InvolIndexPol_Map* hsigmas, const CoxNbr &x, const CoxNbr &sx, const Generator & s) const
{
	const List<CoxNbr>& interval = Get_LowerBruhatInterval_in_W(s,sx);
	const List<kl::KLCoeff>& mu_coeffs = Get_LowerBruhatInterval_in_W_MuCoeffs(s,sx);
	
	for(int i=0; i<interval.size(); i++) 
	{
		CoxNbr z = interval[i];
		kl::KLCoeff mu = mu_coeffs[i];
		
		if(z != x) Combine_HalfPolynomial_Maps(next, hsigmas[z], -mu);
	}	
}










/*
 * Function: Write_hsigma_Array
 * ---------------------------------------------------------
 * Helper function which writes the array of computed polynomials
 * hsigma_{x,y;z}, for a fixed value of of y, to a designated text file.
 *
 */
void TwistedInvolutionsPoset::Write_hsigma_Array(int y_index, InvolIndexPol_Map* array) const
{	
	FILE *output = Get_hsigma_StructureConstants_OutputFile(y_index);
	
	for(int i=0; i<W->order(); i++) 
	{
		InvolIndexPol_Map next = array[i];
		
		for(int j=0; j<size(); j++) 
		{
			if(next.find(j) != next.end())
			{
				SafePolynomial pol = array[i][j];
				pol.PrintCoefficients(output,i,j);
			}
		}
	}
	
	fclose(output);
}



/*
 * Function: InitializeFiles
 * ---------------------------------------------------------
 * Given a vector of file names, this function open and closes
 * each file, so that the the file is left as a blank text file.
 *
 */
void InitializeFiles(const std::vector<string>& outputfiles)
{
	for(int i=0; i<outputfiles.size(); i++)
	{
		FILE *file = fopen(outputfiles[i].c_str(),"w");
		fclose(file);
	}
}



/*
 * Function: DeleteFiles
 * ---------------------------------------------------------
 * Deletes list of files. Should only be called if absolutely
 * sure the temporary files are no longer needed! 
 *
 */
void DeleteFiles(const std::vector<string>& files) 
{
	for(int x = 0; x < files.size(); x++) 
	{
		remove( files[x].c_str() );
	}
}



/*
 * Function: TransposeFiles
 * ---------------------------------------------------------
 * The function takes as input two lists of file names: 
 *
 *   input_x  for  x = 0,1,2,...
 *   output_y for  y = 0,1,2,...
 *
 * File input_x is assumed to be a text file each of whose  
 * lines have the form
 *
 *   y,z;a,b,c,...
 *
 * Here the letters are each integers, and the part after the 
 * semicolon represents some polynomial a + bx + cx^2 + ...
 *
 * The function reads all the input files and appends the line 
 * "x,z;a,b,c,..." to the  output file output_y for each corresponding
 * line "y,z;a,b,c,..." in the input file output_x.
 *
 * After this is carried out, all the input files are deleted.
 *
 */
void TwistedInvolutionsPoset::TransposeFiles(const std::vector<string>& inputfiles, const std::vector<string>& outputfiles) const
{	
	InitializeFiles(outputfiles);
	TransposeData_t data(outputfiles.size());
	
	for(int x = 0; x < inputfiles.size(); x++) 
	{
		PrintAllProgress("  ", x, inputfiles.size());
		
		// Reads data from file input_x
		data.Setup(inputfiles[x]);		
		
		// Writes transposed data to collection of files output_y
		for(int y = 0; y < outputfiles.size(); y++)
		{
			FILE* file = fopen(outputfiles[y].c_str(), "a");
			data.PrintEntries(file, x, y);
		}
		
		data.yz_pairs.clear();
		data.polynomials.clear();
	}

	
	// DeleteFiles(inputfiles);
}



/*
 * Function: Transpose_h_Files
 * ---------------------------------------------------------
 * When the program first computes the structure constants h_{x,y;z},
 * the polynomials are divided into separate files according to
 * their y index. 
 *
 * This function transposes this data so that the polynomials are
 * instead divided into separate files according to their x index.
 *
 * This format make it easier to compute the polynomials htilde_{x,y;z}.
 *
 */
void TwistedInvolutionsPoset::Transpose_h_Files() const
{
	cout << " -Printing polynomials h_{x,y;z}." << endl;
	
	
	std::vector<string> inputfiles;
	for(int y=0; y<size(); y++) inputfiles.push_back( Get_h_StructureConstants_Temp_FileName(y) );
	
	std::vector<string> outputfiles;
	for(int x=0; x<W->order(); x++) outputfiles.push_back( Get_h_StructureConstants_FileName(x) );
	
	
	TransposeFiles(inputfiles, outputfiles);
}



/*
 * Function: Transpose_htilde_Files
 * ---------------------------------------------------------
 * When the program first computes the polynomials htilde_{x,y;z},
 * the polynomials are divided into separate files according to
 * their x index. 
 *
 * This function transposes this data so that the polynomials are
 * instead divided into separate files according to their y index. 
 *
 * This format make it easier to compare the polynomials 
 * htilde_{x,y;z} and hsigma_{x,y;z}
 *
 */
void TwistedInvolutionsPoset::Transpose_htilde_Files() const
{
	cout << " -Printing polynomials htilde_{x,y;z}." << endl;
	
	
	std::vector<string> inputfiles;
	for(int x=0; x<W->order(); x++) inputfiles.push_back( Get_htilde_StructureConstants_Temp_FileName(x) );
	
	std::vector<string> outputfiles;
	for(int y=0; y<size(); y++) outputfiles.push_back( Get_htilde_StructureConstants_FileName(y) );
	
	
	TransposeFiles(inputfiles, outputfiles);
}







/*
 * Function: Print_Comparison
 * ---------------------------------------------------------
 * Helper function to print output message both to file and 
 * command line.
 *
 */
void Print_Comparison(FILE* file, stringstream &output)
{	
	fputs(output.str().c_str(), file);
	fclose(file);
	
	cout << endl << endl; 
	cout << "*************" << endl;
	cout << "** SUMMARY **" << endl;
	cout << "*************" << endl << output.str();
}





/*
 * Function: GetNormalForm
 * ---------------------------------------------------------
 * Returns a string giving reduced expression for the element
 * in W represented by the CoxNbr y. 
 *
 */
string TwistedInvolutionsPoset::GetNormalForm(CoxNbr y) const
{
	if(y == 0) return "";
	
	Generator s = firstBit(W->ldescent(y));
	W->lprod(y,s);
	
	stringstream ss;
	ss << (s+1) << GetNormalForm(y);
	
	return ss.str();
}



/*
 * Function: PrintElement
 * ---------------------------------------------------------
 * One line method called in function PrintCoefficientStatistics_Helper
 *
 */
void PrintElement(stringstream &ss, char x, string normal_form)
{
	ss << "  " << x << " = " << normal_form << endl;
}
	


/*
 * Function: PrintCoefficientStatistics_Helper
 * ---------------------------------------------------------
 * Helper function called by PrintCoefficientStatistics.
 *
 */
void TwistedInvolutionsPoset::PrintCoefficientStatistics_Helper(stringstream &ss, ComparisonData_t &d, string size_adj, SafeCoeff & coeff, SafePolynomial & pol, std::vector<int> &index) const
{
	string pol_str;
	if(d.isPrintingStructureConstants())
	{
		pol_str = pol.toString("v",pol.Degree()/2);
	}
	else 
	{
		pol_str = pol.toString();
	}
		
	
	ss << size_adj << " coefficient of any " << d.tag << ": " << coeff << endl;
	ss << "Exhibited by:" << endl;
	ss << endl;
	ss << "  " << d.tag << " = " << pol_str << endl;
	ss << endl;
	
	
	// Prints the indices attached to the polynomial pol: either x,y,z or y,w
	//
	if(d.isPrintingStructureConstants())
	{
		PrintElement(ss, d.symbols[0], GetNormalForm(index[0]));
		PrintElement(ss, d.symbols[1], GetNormalForm(involutions[index[1]]));
		PrintElement(ss, d.symbols[2], GetNormalForm(involutions[index[2]]));
	}
	else
	{
		PrintElement(ss, d.symbols[0], GetNormalForm(involutions[index[0]]));
		PrintElement(ss, d.symbols[1], GetNormalForm(involutions[index[1]]));
	}

	ss << endl;
}



/*
 * Function: PrintCoefficientStatistics
 * ---------------------------------------------------------
 * Given a ComparisonData_t d as input, print to stringstream passed
 * as a parameter the smallest and largest coefficients occuring among
 * the polynomials stored in d.
 *
 */
void TwistedInvolutionsPoset::PrintCoefficientStatistics(stringstream &ss, ComparisonData_t &d) const
{
	if(d.mincoeff_index.size() == 0 || d.maxcoeff_index.size() == 0)
	{
		ss << "(All polynomials are zero)" << endl;
		ss << endl;
	}
	else
	{
		PrintCoefficientStatistics_Helper(ss, d, "Smallest", d.mincoeff, d.mincoeff_pol, d.mincoeff_index);
		PrintCoefficientStatistics_Helper(ss, d, "Largest", d.maxcoeff, d.maxcoeff_pol, d.maxcoeff_index);
	}
	
	// If d stores P's and not h's then we keep track of the number of distinct polynomials
	// and print this number here. For structure constants h, there are too many distinct
	// polynomials (at least in type H4) to be kept track of during the computation.
	//
	if(! d.isPrintingStructureConstants()) ss << "Number of distinct polynomials " << d.tag << ": " << d.size() << endl;
}



/*
 * Function: Append_Group_Data
 * ---------------------------------------------------------
 * Prints to input stringstream the type of the current Coxeter system
 * along with the choice of involution * and the sizes of W and I_*.
 *
 */
void TwistedInvolutionsPoset::Append_Group_Data(stringstream &ss) const
{
	ss << endl;
	ss << "Type of Coxeter system (W,S): " << group_name.substr(0,group_name.size()-W->rank()-1) << endl;
	ss << "Choice of involution *      : " << group_name.substr(group_name.size()-W->rank()) << endl;
	ss << endl;
	ss << "Size of W  : " << W->order() << endl;
	ss << "Size of I_*: " << size() << endl;
	ss << endl;	
	ss << endl;
}



/*
 * Function: Append_Comparison_Output
 * ---------------------------------------------------------
 * Prints to input stringstream various summary data stored
 * in the input ComparisonData_t parameter.
 *
 */
void TwistedInvolutionsPoset::Append_Comparison_Output(stringstream &ss, ComparisonData_t &d) const
{
	ss << endl;
	
	// Prints boolean properties of polynomial family stored in d
	d.PrintProperties(ss);
	
	ss << endl;
	
	// Prints greatest and smallest coefficients of polynomial family
	PrintCoefficientStatistics(ss,d);	
	
	ss << endl;
	ss << endl;	
}



/*
 * Function: GetNextPair
 * ---------------------------------------------------------
 * Given input polynomials X and X_sigma, constructs 
 *
 *   (1/2) * (X + X_sigma)		and		(1/2) * (X - X_sigma)
 *
 * and stores these polynomials in the parameters X_plus and 
 * X_neg passed by reference.
 *
 */
void GetNextPair(SafePolynomial & X_plus, SafePolynomial & X_neg, const SafePolynomial &X, const SafePolynomial &X_sigma)
{
	X_plus = X;
	X_plus += X_sigma;
	X_plus.Divide_By_Two();
	
	X_neg = X;
	X_neg -= X_sigma;
	X_neg.Divide_By_Two();
}

void GetNextPair(std::set<std::pair<int,int > >::const_iterator &it, ComparisonData_t & X_plus, ComparisonData_t & X_neg, ComparisonData_t &X, ComparisonData_t &X_sigma)
{
	GetNextPair(X_plus.tmp_list[*it], X_neg.tmp_list[*it], X.tmp_list[*it], X_sigma.tmp_list[*it]);
}



/*
 * Function: Test_Polynomials_Helper
 * ---------------------------------------------------------
 * Helper function called by Test_Polynomials. Reads the polynomials
 * P_{y,w}, Psigma_{y,w}, P^+_{y,w}, P^-_{y,w} from text files
 * and stores this data in four ComparisonData_t objects.
 *
 */
void TwistedInvolutionsPoset::Test_Polynomials_Helper(ComparisonData_t &P, ComparisonData_t &P_sigma, ComparisonData_t &P_plus, ComparisonData_t &P_neg) const
{
	std::set<std::pair<int,int > > keys;
	
	// Constructs lists of polynomials P_{y,w} and Psigma_{y,w}
	SafePolynomial::ReadFilePolynomials(Get_P_Polynomials_FileName(), keys, P.tmp_list);
	SafePolynomial::ReadFilePolynomials(Get_Psigma_Polynomials_FileName(), keys, P_sigma.tmp_list);
	
	// Constructs lists of polynomials P^+_{y,w} and P^-_{y,w}
	std::set<std::pair<int,int > >::const_iterator it;
	for(it = keys.begin(); it != keys.end(); ++it)
	{
		GetNextPair(it, P_plus, P_neg, P, P_sigma);
	}
	
	P.UpdatePolynomials(bruhat_ascents);	
	P_sigma.UpdatePolynomials(bruhat_ascents);	
	P_plus.UpdatePolynomials(bruhat_ascents);	
	P_neg.UpdatePolynomials(bruhat_ascents);
}



/*
 * Function: Test_Polynomials
 * ---------------------------------------------------------
 * Reads the polynomials P_{y,w} and Psigma_{y,w} from their
 * files in the output directory, assuming they have already
 * been computed, and checks several properties of the four 
 * families of polynomials
 *
 *   P_{y,w}
 *   Psigma_{y,w}
 *   P^+_{y,w} := (1/2) * [ P_{y,w} + Psigma_{y,w} ]
 *   P^-_{y,w} := (1/2) * [ P_{y,w} - Psigma_{y,w} ]
 *
 * Precisely which properties are checked is controlled by the 
 * method PrintProperties_of_Polynomials in "CoxeterExtensions_by_EM_Structs.cpp".
 *
 * The outcome of this analysis is printed to a text file
 * in the folder labeled OUTPUT_FOLDER in the output directory.
 *
 */
void TwistedInvolutionsPoset::Test_Polynomials() const
{
	char y = 'y', w = 'w';
	
	// Intializes ComparisonData_t's 
	ComparisonData_t P		 ("P_{y,w}", y, w);
	ComparisonData_t P_sigma ("Psigma_{y,w}", y, w);
	ComparisonData_t P_plus  ("P^+_{y,w}", y, w);
	ComparisonData_t P_neg	 ("P^-_{y,w}", y, w);

	// Imports polynomials from output text files to ComparisonData_t's
	Test_Polynomials_Helper(P,P_sigma,P_plus,P_neg);
	
	stringstream output;
	
	Append_Group_Data(output);
	
	// Prints to output the notable statistics from ComparisonData_t's
	Append_Comparison_Output(output, P);
	Append_Comparison_Output(output, P_sigma);
	Append_Comparison_Output(output, P_plus);
	Append_Comparison_Output(output, P_neg);
	
	// Prints output to text file and to cout
	Print_Comparison(Get_Summary_OutputFile(), output);
}



/*
 * Function: Polynomial_to_ShiftedLaurentPolynomial
 * ---------------------------------------------------------
 * Replaces the input polynomial 
 *
 *   f = a + bx + cx^2 + ... + dx^n
 *
 * with 
 *
 *   f = d + ... + cx^{n-2} + bx^{n-1} + ax^n + bx^{n+1} + cx^{n+2} + ... + dx^{2n}
 *
 */
void Polynomial_to_ShiftedLaurentPolynomial(SafePolynomial& f)
{
	if(f.isZero()) return;
	
	int d = f.Degree();
	
	f *= SafePolynomial(d);
	for(int i=1; i<=d; i++) f += SafePolynomial(d-i,f[d+i]);
}



/*
 * Function: Test_StructureConstants_Helper
 * ---------------------------------------------------------
 * Helper function called by Test_StructureConstants. Reads in the Laurent 
 * polynomials htilde_{x,y;z}, hsigma_{x,y;z}, h^+_{x,y;z}, h^-_{x,y;z} 
 * with y fixed, and stores this data in four ComparisonData_t objects.
 *
 * The Laurent polynomial ax^-d + ... + bx^d is stored here as the
 * Safepolynomial object a + ... + bx^2d.
 *
 */
void TwistedInvolutionsPoset::Test_StructureConstants_Helper(int y, ComparisonData_t &h_tilde, ComparisonData_t &h_sigma, ComparisonData_t &h_plus, ComparisonData_t &h_neg) const
{
	std::set<std::pair<int,int > > keys;

	// Constructs lists of polynomials htilde_{x,y;z} and hsigma_{x,y;z}
	SafePolynomial::ReadFilePolynomials(Get_htilde_StructureConstants_FileName(y), keys, h_tilde.tmp_list);
	SafePolynomial::ReadFilePolynomials(Get_hsigma_StructureConstants_FileName(y), keys, h_sigma.tmp_list);
	
	// Constructs lists of polynomials h^+_{x,y;z} and h^-_{x,y;z}
	std::set<std::pair<int,int > >::const_iterator it;
	for(it = keys.begin(); it != keys.end(); ++it)
	{			
		GetNextPair(it, h_plus, h_neg, h_tilde, h_sigma);
		
		// Converts ax^-d + ... + bx^d to a + ... + bx^2d
		Polynomial_to_ShiftedLaurentPolynomial(h_tilde.tmp_list[*it]);
		Polynomial_to_ShiftedLaurentPolynomial(h_sigma.tmp_list[*it]);
		Polynomial_to_ShiftedLaurentPolynomial(h_plus.tmp_list[*it]);
		Polynomial_to_ShiftedLaurentPolynomial(h_neg.tmp_list[*it]);
	}
	
	h_tilde.UpdateStructureConstants(y);
	h_sigma.UpdateStructureConstants(y);
	h_plus.UpdateStructureConstants(y);
	h_neg.UpdateStructureConstants(y);	
}



/*
 * Function: Test_StructureConstants
 * ---------------------------------------------------------
 * Reads the polynomials htilde_{x,y;z} and hsigma_{x,y;z} from their
 * files in the output directory, assuming they have already
 * been computed, and checks several properties of the four families
 * of polynomials
 *
 *   htilde_{x,y;z}
 *   hsigma_{x,y;z}
 *   h^+_{x,y;z} := (1/2) * [ htilde_{x,y;z} + hsigma_{x,y;z} ]
 *   h^-_{x,y;z} := (1/2) * [ htilde_{x,y;z} 0 hsigma_{x,y;z} ]
 *
 * Precisely which properties are checked is controlled by the method
 * CheckProperties_of_StructureConstants in "CoxeterExtensions_by_EM_Structs.cpp".
 
 * The outcome of this analysis is printed to a text file
 * in the folder labeled OUTPUT_FOLDER in the output directory.
 *
 */
void TwistedInvolutionsPoset::Test_StructureConstants() const
{
	cout << "Analyzing polynomials h^{+/-}_{x,y;z}." << endl;
	
	char x = 'x', y = 'y', z = 'z';
	
	// Intializes ComparisonData_t's 
	ComparisonData_t h_tilde ("htilde_{x,y;z}", x, y, z);
	ComparisonData_t h_sigma ("hsigma_{x,y;z}", x, y, z);
	ComparisonData_t h_plus  ("h^+_{x,y;z}", x, y, z);
	ComparisonData_t h_neg	 ("h^-_{x,y;z}", x, y, z);
	
	// Imports polynomials from output text files to ComparisonData_t's
	for(int y_index = 0; y_index < size(); y_index++)
	{
		PrintAllProgress("  ", y_index, size());
		Test_StructureConstants_Helper(y_index, h_tilde, h_sigma, h_plus, h_neg);
	}
	
	stringstream output;
	
	// Prints to output the notable statistics from ComparisonData_t's
	Append_Comparison_Output(output, h_tilde);
	Append_Comparison_Output(output, h_sigma);
	Append_Comparison_Output(output, h_plus);
	Append_Comparison_Output(output, h_neg);
	
	// Prints output to text file and to cout
	Print_Comparison(Get_Summary_OutputFile("a"), output);
	
	cout << "Output printed to \"" << Get_Summary_FileName() << "\"" << endl << endl;
}




/*
 * GetFileName functions
 * ---------------------------------------------------------
 * The following functions provide the names of various output
 * files. Each file name has the form
 *
 *   (CoxeterType)_(Involution)_(Label)_(OptionalIndex)
 *
 * For example, "F4_4321_hsigma_StructureConstants_15"
 *
 */
string TwistedInvolutionsPoset::GetFileName(const string &folder, const string &file) const
{
	stringstream ss;
	ss << folder << group_name << "_" << file;
	return ss.str();
}

string TwistedInvolutionsPoset::GetFileName(const string &folder, const string &file, int i) const
{
	string name = GetFileName(folder,file);	
	
	stringstream ss;
	ss << name << "_" << i;
	return ss.str();
}

string TwistedInvolutionsPoset::Get_InvolutionIndices_FileName() const {
	return GetFileName(INDICES_INVOLUTIONS_FOLDER, INDICES_INVOLUTIONS_LABEL);
}

string TwistedInvolutionsPoset::Get_AllElementIndices_FileName() const {
	return GetFileName(INDICES_ALL_ELEMENTS_FOLDER, INDICES_ALL_ELEMENTS_LABEL);
}

string TwistedInvolutionsPoset::Get_P_Polynomials_FileName() const {
	return GetFileName(POLYNOMIALS_P_FOLDER, POLYNOMIALS_P_LABEL);
}

string TwistedInvolutionsPoset::Get_Psigma_Polynomials_FileName() const {
	return GetFileName(POLYNOMIALS_Psigma_FOLDER, POLYNOMIALS_Psigma_LABEL);
}

string TwistedInvolutionsPoset::Get_h_StructureConstants_FileName(int x_index) const {
	return GetFileName(STRUCTURECONSTANTS_h_FOLDER, STRUCTURECONSTANTS_h_LABEL, x_index);
}

string TwistedInvolutionsPoset::Get_h_StructureConstants_Temp_FileName(int y_index) const {
	return GetFileName(STRUCTURECONSTANTS_h_TEMP_FOLDER, STRUCTURECONSTANTS_h_TEMP_LABEL, y_index);
}

string TwistedInvolutionsPoset::Get_htilde_StructureConstants_FileName(int y_index) const {
	return GetFileName(STRUCTURECONSTANTS_htilde_FOLDER, STRUCTURECONSTANTS_htilde_LABEL, y_index);
}

string TwistedInvolutionsPoset::Get_htilde_StructureConstants_Temp_FileName(int x_index) const {
	return GetFileName(STRUCTURECONSTANTS_htilde_TEMP_FOLDER, STRUCTURECONSTANTS_htilde_TEMP_LABEL, x_index);
}

string TwistedInvolutionsPoset::Get_hsigma_StructureConstants_FileName(int y_index) const {
	return GetFileName(STRUCTURECONSTANTS_hsigma_FOLDER, STRUCTURECONSTANTS_hsigma_LABEL, y_index);
}

string TwistedInvolutionsPoset::Get_Summary_FileName() const {
	return GetFileName(OUTPUT_FOLDER, Summary_LABEL);
}

FILE* TwistedInvolutionsPoset::Get_P_Polynomials_OutputFile() const {
	return fopen(Get_P_Polynomials_FileName().c_str(),"w");
}

FILE* TwistedInvolutionsPoset::Get_Psigma_Polynomials_OutputFile() const {
	return fopen(Get_Psigma_Polynomials_FileName().c_str(),"w");
}

FILE* TwistedInvolutionsPoset::Get_h_StructureConstants_TempFile(int y_index) const {
	return fopen(Get_h_StructureConstants_Temp_FileName(y_index).c_str(), "w");
}

FILE* TwistedInvolutionsPoset::Get_htilde_StructureConstants_TempFile(int x_index) const {
	return fopen(Get_htilde_StructureConstants_Temp_FileName(x_index).c_str(), "w");
}

FILE* TwistedInvolutionsPoset::Get_hsigma_StructureConstants_OutputFile(int y_index) const {
	return fopen(Get_hsigma_StructureConstants_FileName(y_index).c_str(), "w");
}

FILE* TwistedInvolutionsPoset::Get_Summary_OutputFile(const char * write_type) const {
	return fopen(Get_Summary_FileName().c_str(), write_type);
}





