
/* 
 * File: "CoxeterExtensions_by_EM_SafePolynomial.cpp"
 *
 * Author: Eric Marberg
 * Version: June 2013
 * --------------------------------------------------------------
 * Source file for "CoxeterExtensions_by_EM_SafePolynomial.h"
 *
 */



#include "CoxeterExtensions_by_EM_SafePolynomial.h"



// This line enables code like "cout << pol << endl;" where pol is a SafePolynomial
ostream& operator<<(ostream& out, const SafePolynomial& p) { return out << p.toString(); }



/*
 * Function: SafePolynomial constructor
 * ---------------------------------------------------------
 * Creates the SafePolynomial c*x^d. The parameter c is set to 1 by default.
 *
 */
SafePolynomial::SafePolynomial (int d, SafeCoeff c) 
{
	if(c != 0) 
	{
		for(int i=0; i<d; i++) coefficients.push_back(0);
		coefficients.push_back(c);
	}
}



/*
 * Function: toString
 * ---------------------------------------------------------
 * Converts SafePolynomial to friendly string format for command 
 * line output, to aid in debugging, etc.
 *
 * The optional parameter tag specifies the string which is the
 * variable of the polynomial; this is "q" by default.
 *
 * If the second optional parameter power_shift is set to something
 * nonzero (it is zero by default), then the function returns the 
 * string representation of (*this) * x^{-power_shift}.
 *
 */
string SafePolynomial::toString(const string &tag, int power_shift) const {
	
	if(coefficients.empty()) return "0";
	
	stringstream ss;
	bool first = true;
	
	for(int p=0; p<coefficients.size(); p++) {
		
		SafeCoeff coeff = coefficients[p];
		int power = p - power_shift;
		
		// Skips to next iteration if coefficient is zero
		if(coeff == 0) continue;
		
		
		// Determines prefix " + ", " - ", or "-" for next monomial
		// 
		if(first) {
			if(coeff < 0) {
				ss << "-";
				coeff = -coeff;
			}
		} 
		else {
			if(coeff < 0) {
				ss << " - ";
				coeff = -coeff;
			} else {
				ss << " + ";
			}
		}
		
		first = false;
		
		// Determine how to print monomial, e.g. as "3x^2" or "1" or "x"
		//
		if(coeff == 1) {
			if(power == 0) ss << coeff;
			if(power == 1) ss << tag;
			if(power != 0 && power != 1) ss << tag << "^" << power;
		} else {
			if(power == 0) ss << coeff;
			if(power == 1) ss << coeff << tag;
			if(power != 0 && power != 1) ss << coeff << tag << "^" << power;			
		}	
	}
	
	return ss.str();
}



/*
 * Function: GetNonzeroCoefficients
 * ---------------------------------------------------------
 * Returns vector of nonzero coefficients, represented as 
 * constant SafePolynomials.
 *
 */
std::vector<SafePolynomial> SafePolynomial::GetNonzeroCoefficients() const
{
	std::vector<SafePolynomial> nonzero_coeffs; 
	for(int i = 0; i <= Degree(); i++) 
	{
		if((*this)[i] != 0) 
		{
			nonzero_coeffs.push_back(SafePolynomial(0,(*this)[i]));
		}
	}
	
	return nonzero_coeffs;
}



/*
 * Function: Print
 * ---------------------------------------------------------
 * Calls PrintList on polynomial's vector of coefficients.
 *
 */
void SafePolynomial::PrintCoefficients(FILE *output, int i, int j) const
{
	if(isZero()) return;
	
	PrintList(output, i, j, coefficients, SafeCoeff_TAG);
}



/*
 * Function: Divide_By_Two
 * ---------------------------------------------------------
 * Modifies the SafePolynomial by replacing each of its coefficients
 * c with c / 2. Announces an error and aborts the program if any c
 * is odd.
 *
 */
void SafePolynomial::Divide_By_Two()
{
	for(int i=0; i<coefficients.size(); i++) 
	{
		if(coefficients[i] % 2 != 0) 
		{
			AnnounceErrorMessage("Division error; cannot divide by two:\n\n" + toString() + "\n");
		}
		coefficients[i] /= 2;
	}
}



/*
 * Function: SafePolynomial operator[]
 * ---------------------------------------------------------
 * Returns jth coefficient of SafePolynomial, or zero if j is
 * less than 0 or greater than the degree.
 *
 */
const SafeCoeff SafePolynomial::operator[] (int j) const 
{ 
	if(j>=0 && j<coefficients.size()) return coefficients[j]; 
	else return 0;
}



/*
 * Function: SafePolynomial operator<
 * ---------------------------------------------------------
 * Returns true if the highest degree coefficient of (other - this)
 * is positive, false otherwise. Induces total order of SafePolynomials.
 *
 */
bool SafePolynomial::operator<(const SafePolynomial &other) const
{
	SafePolynomial diff = other;
	diff -= *this;
	
	return !diff.isZero() && diff[diff.Degree()] > 0;
}



/*
 * Function: SafePolynomial operator=
 * ---------------------------------------------------------
 * Returns true iff the difference between this and other is zero.
 *
 */
SafePolynomial& SafePolynomial::operator=(const SafePolynomial &param)
{
	coefficients = param.coefficients;
	return *this;
}	



/*
 * Function: SafePolynomial operator*=
 * ---------------------------------------------------------
 * Operator for multiplying two polynomials. The method carefully
 * checks for integer overflow errors after all intermediate 
 * computations.
 *
 */
SafePolynomial& SafePolynomial::operator*=(const SafePolynomial &param) 
{	
	if(param.isZero() || isZero()) coefficients.clear();
	else
	{
		std::vector<SafeCoeff> new_coeffs = coefficients;
		for(int i=0; i<=Degree(); i++) SafeMultiply(new_coeffs[i],param[0]);
		
		for(int i=1; i<=param.Degree(); i++)
		{
			new_coeffs.push_back(0);
			if(param[i] != 0)
			{
				for(int j=0; j<=Degree(); j++)
				{
					SafeCoeff x = param[i];
					SafeMultiply(x, coefficients[j]);
					SafeAdd(new_coeffs[i+j], x);
				}
			}
		}
		
		coefficients = new_coeffs;
	}
	
	return *this;
}



/*
 * Function: ShiftDown
 * ---------------------------------------------------------
 * Deletes monomials 1,x,x^2,...,x^{n-1} from current polynomial 
 * then divides the result by x^n
 *
 */
void SafePolynomial::ShiftDown(int n) 
{ 
	if(n >= coefficients.size()) coefficients.clear();
	else coefficients.erase(coefficients.begin(),coefficients.begin()+n); 
}



/*
 * Function: SafeAdd
 * ---------------------------------------------------------
 * Static method for adding two SafeCoeffs. The function does
 * the same thing as "a += b" except that it checks for
 * integer overflow and aborts the program if overflow occurs.
 *
 */
void SafePolynomial::SafeAdd(SafeCoeff &a, const SafeCoeff &b)
{
	bool a_leq_zero = a <= 0;
	
	a += b;
	
	if( (a_leq_zero && a > b) || (!a_leq_zero && a <= b) ) 
	{
		AnnounceOverflowError(a-b,b,'+');
	}
}



/*
 * Function: SafeMultiply
 * ---------------------------------------------------------
 * Static method for multiplying two SafeCoeffs. The function does
 * the same thing as "a *= b" except that it checks for integer
 * overflow and aborts the program if overflow occurs.
 *
 */
void SafePolynomial::SafeMultiply(SafeCoeff &a, const SafeCoeff &b)
{	
	if(b != 0 && abs(a) > SafeCoeff_MAX / abs(b)) 
	{
		AnnounceOverflowError(a,b,'*');
	}
	
	a *= b;
}



/*
 * Function: SafeAddHelper
 * ---------------------------------------------------------
 * Helper function for addition / subtraction of SafePolynomials.
 * The code 
 *
 *   (*this).SafeAddHelper(param,add)
 *
 * is equivalent to (*this) += param if add == true and to 
 * (*this) -= param otherwise.
 *
 */
SafePolynomial& SafePolynomial::SafeAddHelper(const SafePolynomial &param, bool add) 
{
	int d = Degree();
	int e = param.Degree();
	
	for(int i = d+1 ; i <= e; i++) 
	{
		coefficients.push_back( add ? param[i] : -param[i] );
	}
	
	for(int i = 0; i <= std::min(d,e); i++) 
	{
		SafeAdd(coefficients[i], add ? param[i] : -param[i]);
	}
	
	while(coefficients.size() > 0 && coefficients.back() == 0) 
	{
		coefficients.pop_back();
	}
	
	return *this;
}



/*
 * Function: MultiplyHalfPolynomials
 * ---------------------------------------------------------
 * Given a polynomial f = a + bx + cx^2 + ... define Double(f) 
 * as the Laurent polynomial
 *
 *   Double(f) = a + b(x + 1/x) + c(x^2 + 1/x^2) + ...
 *
 * Likewise given a Laurent polynomial g = ... + a/x + b + cx + ...
 * define Halve(g) as the polynomial 
 *
 *   Halve(g) = b + cx + ...
 *
 * This static method computes and returns the result of
 *
 *   Halve( Double(F) * Double(G) ),
 * 
 * checking for overflow errors and aborting if any occur.
 *
 */
SafePolynomial SafePolynomial::MultiplyHalfPolynomials(const SafePolynomial &F, const SafePolynomial &G)
{
	if(F.isZero() || G.isZero()) return SafePolynomial::GetZero();
	
	SafePolynomial ans = F;
	ans *= G[0];
	
	for(int i=1; i<= G.Degree(); i++)
	{
		if(G[i] != 0) MultiplyHalfPolynomials_Helper(ans,F,G[i],i);
	}
	
	return ans;
}



/*
 * Function: MultiplyHalfPolynomials_Helper
 * ---------------------------------------------------------
 * Static helper method used in MultiplyHalfPolynomials.
 *
 */
void SafePolynomial::MultiplyHalfPolynomials_Helper(SafePolynomial& ans, const SafePolynomial &f, SafeCoeff coeff, int degree)
{
	// Let d = degree and f = a_0 + a_1x + ... + a_nx^n
	
	SafePolynomial up = f;
	up *= SafePolynomial(degree);
	for(int i=0; i<degree; i++) up += SafePolynomial(i,f[degree-i]);
	
	// Now up = a_d + a_{d-1}x + a_{d-2}x^2 + ... + a_0x^d + a_1x^{d+1} + ... + a_nx^{d+n}
	
	SafePolynomial down = f;
	down.ShiftDown(degree);
	
	// Now down = a_d + a_{d+1}x + a_{d+2}x^2 + ... + a_nx^{n-d}
	
	// Replaces up by (up + down) * coeff
	up += down;
	up *= coeff;
	
	// Adds up to ans
	ans += up;
}



/*
 * Function: AnnounceErrorMessage
 * ---------------------------------------------------------
 * Static method which in case of integer overflow or some 
 * other fatal error, prints an error message and causes  
 * program to terminate. 
 *
 * Not most robust error handling, but ensures that overflow
 * errors are caught.
 *
 */
void SafePolynomial::AnnounceErrorMessage(const string &message)
{
	cout << endl << "*** ERROR ***" << endl << endl << message << endl;
	cout << endl << "*** Exiting program ***" << endl << endl;
	
	exit(1);
}



/*
 * Function: AnnounceOverflowError
 * ---------------------------------------------------------
 * Static method which Announces overflow error from either 
 * addition or multiplication.
 *
 */
void SafePolynomial::AnnounceOverflowError(const SafeCoeff &a, const SafeCoeff &b, char symbol)
{
	stringstream ss;
	
	if(symbol == '+') ss << "-Addition overflow: a + b";
	else ss << "-Multiplication overflow: a * b";
	
	ss << " with a = " << a << " and b = " << b;
	AnnounceErrorMessage(ss.str());
}



/*
 * Function: SplitLine
 * ---------------------------------------------------------
 * Return the vector of strings
 *
 *   s1, s2, s3, ..., sn
 *
 * such that the input string has the form
 *
 *   line = s1 + delim + s2 + delim + s3 + delim + ... + delim + sn
 *
 */
std::vector<string> SplitLine (string line, const string & delim) 
{
	std::vector<string> output;
	
	size_t pos = 0;
	while(pos != string::npos)
	{ 
		pos = line.find(delim, 0); 
		
		string next;
		if(pos == string::npos) next = line.substr(0);
		else next = line.substr(0, pos);
		
		output.push_back(line.substr(0, pos));		
		line = line.substr(pos+1);
	}
	
	return output;
}



/*
 * Function: ReadSinglePolynomial
 * ---------------------------------------------------------
 * Reads a single line from the given input text file. If
 * the line is the empty string "" then the function returns
 * false to indicate that we have reached the end of the input
 * file. Otherwise, we assume the line of text has the form
 * 
 *   i,j;a,b,c,d,...
 *
 * where i,j are ints and a,b,c,d,... are strings representing 
 * longs (or empty strings "" which are interpreted as "0"). 
 * In this case the function sets 
 *
 *   y_index = i
 *   z_index = j 
 *   pol = a + bx + cx^2 + dx^3 + ...
 *
 * and returns true, to indicate that we have not yet reached 
 * the end of the input file.
 *
 */
bool SafePolynomial::ReadSinglePolynomial(ifstream &file, int &y_index, int &z_index, SafePolynomial &pol)
{	
	string line;
	getline (file,line); 
	
	if(line == "") return false;
	else 
	{
		std::vector<string> line_parts = SplitLine(line,";");
		
		std::vector<string> indices = SplitLine(line_parts[0],",");
		y_index = atoi(indices[0].c_str());
		z_index = atoi(indices[1].c_str());
		
		std::vector<string> coeffs_str = SplitLine(line_parts[1],",");
		std::vector<SafeCoeff> coeffs;
		for(int i=0; i<coeffs_str.size(); i++) 
		{
			SafeCoeff c = 0;
			
			// empty string indicates that c = 0
			// otherwise we parse the integer value from string
			if(coeffs_str[i] != "")
			{
				c = strtol(coeffs_str[i].c_str(), NULL, SafeCoeff_BASE);
				
				// Checks that we were able to parse value correctly
				if(c == SafeCoeff_MAX || c == SafeCoeff_MIN) 
				{
					AnnounceErrorMessage("Coefficient out of range for SafeCoeff type");
				}
			}
			
			coeffs.push_back(c);
		}
		
		pol = SafePolynomial(coeffs);
	}
	
	return true;
}




/*
 * Function: ReadFilePolynomials
 * ---------------------------------------------------------
 * Given a text file specified by the string filename, each of
 * whose lines has the format
 *
 *   i,j;a,b,c,d...
 *
 * this function places the pair <i,j> in the set keys and the
 * polynomial a+bx+cx^2+dx^3+... in the map X with key <i,j>.
 *
 */
void SafePolynomial::ReadFilePolynomials(const string &filename, std::set<std::pair<int,int > > &keys, std::map<std::pair<int,int>, SafePolynomial> &X)
{
	int y,w;
	SafePolynomial p;
	
	ifstream file(filename.c_str());
	while(ReadSinglePolynomial(file,y,w,p))
	{
		std::pair<int,int> pr = make_pair(y,w);
		
		keys.insert(pr);
		X[pr] = p;
	}
	file.close();
}



/*
 * Function: WriteSafePolynomialArray
 * ---------------------------------------------------------
 * Prints each entry in a size-by-size array of SafePolynomials
 * to a given output file. 
 *
 */
void SafePolynomial::WriteSafePolynomialArray(FILE *output, int size, SafePolynomial** array)
{
	for(int i=0; i<size; i++) 
	{
		for(int j=0; j<size; j++) 
		{
			SafePolynomial pol = array[i][j];
			pol.PrintCoefficients(output,i,j);
		}
	}
	
	fclose(output);
}



/*
 * Function: Delete_2D_PolynomialArray
 * ---------------------------------------------------------
 * Frees memory used for size-by-size SafePolynomial**
 *
 */
void SafePolynomial::Delete_2D_PolynomialArray(SafePolynomial **arr, int size)
{
	for(int i=0; i<size; i++) delete[] arr[i];
	delete[] arr;
}



/*
 * Function: ReadSafePolynomialArray
 * ---------------------------------------------------------
 * Given a text file specified by the string filename, each of
 * whose lines has the format
 *
 *   i,j;a,b,c,d...
 *
 * where i and j range from 0 to size-1, this function returns
 * a size-by-size array of SafePolynomials with a+bx+cx^2+...
 * in entry (i,j).
 *
 */
SafePolynomial** SafePolynomial::ReadSafePolynomialArray(const string &filename, int size)
{
	SafePolynomial** arr = InitializeSafePolynomialArray(size);
	
	int i,j;
	SafePolynomial p;
	
	ifstream file( filename.c_str() );
	while(ReadSinglePolynomial(file,i,j,p))
	{
		arr[i][j] = p;
	}
	file.close();
	
	return arr;
}



/*
 * Function: InitializeSafePolynomialArray
 * ---------------------------------------------------------
 * Returns size-by-size array of SafePolynomials, all entries
 * initialized to zero.
 *
 */
SafePolynomial** SafePolynomial::InitializeSafePolynomialArray(int size) 
{
	SafePolynomial** array = new SafePolynomial*[size];
	for(int i=0; i<size; i++) 
	{
		array[i] = new SafePolynomial[size];
		for(int j=0; j<size; j++) 
		{
			array[i][j] = SafePolynomial::GetZero();
		}
	}
	return array;
}	



