
/* 
 * File: "CoxeterExteions_by_EM_SafePolynomial.h"
 *
 * Author: Eric Marberg
 * Version: June 2013
 * --------------------------------------------------------------
 * Header file for SafePolynomial class. Also contains two templatized
 * methods involved with printing polynomials of various kinds to files.
 *
 */



#ifndef __SafePolynomial__
#define __SafePolynomial__



#include "commands.h"
#include "polynomials.h"
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <set>
#include <map>

using namespace std;
using namespace commands;



// Constants associated with type used for coefficients in the SafePolynomial class.
// On my machine, using the data type int for SafeCoeff causes overflow errors for the Coxeter group H4
//
typedef long SafeCoeff;
#define SafeCoeff_MAX LONG_MAX
#define SafeCoeff_MIN LONG_MIN


// Constants involved in printing SafeCoeff to text files
//
#define SafeCoeff_TAG  "%ld"
#define SafeCoeff_BASE 10



/*
 * Class: SafePolynomial
 * ---------------------------------------------------------
 * This class implements a simple univariate polynomial, whose
 * coefficients are SafeCoeffs (i.e., longs), and for which 
 * addition, subtraction, and muliplication are defined to 
 * carefully detect any integer overflow errors.
 *
 * All of the relevant polynomials computed in my extensions to
 * Coxeter are stored as SafePolynomials.
 *
 */
class SafePolynomial {
	
	static const int ZERO_DEGREE = -1;
	
	std::vector<SafeCoeff> coefficients;
	
	static void SafeAdd(SafeCoeff &, const SafeCoeff &);
	static void SafeMultiply(SafeCoeff &, const SafeCoeff &);
	static void MultiplyHalfPolynomials_Helper(SafePolynomial&, const SafePolynomial &, SafeCoeff, int);
	
	SafePolynomial& SafeAddHelper(const SafePolynomial &, bool);
	void ShiftDown(int);
	
public:
	
	SafePolynomial() {}
	~SafePolynomial() {}
	
	SafePolynomial(int, SafeCoeff = 1);
	SafePolynomial(const Polynomial<unsigned> &p)			{ for(int n=0; n<=p.deg(); n++) coefficients.push_back(p[n]); }
	SafePolynomial(const std::vector<SafeCoeff> &coeffs)	{ coefficients = coeffs; }
	
	
	const SafeCoeff operator[] (int) const;	
	SafePolynomial& operator=(const SafePolynomial &);
	
	bool operator<(const SafePolynomial &) const;
	bool operator==(const SafePolynomial &param) const { return coefficients == param.coefficients; }

	SafePolynomial& operator*=(const SafePolynomial &);
	SafePolynomial& operator+=(const SafePolynomial &param)	{ return SafeAddHelper(param,true); }
	SafePolynomial& operator-=(const SafePolynomial &param)	{ return SafeAddHelper(param,false); }
	SafePolynomial& operator*=(SafeCoeff a)					{ return ((*this) *= SafePolynomial(0,a)); }
	SafePolynomial& operator+=(SafeCoeff a)					{ return ((*this) += SafePolynomial(0,a)); }
	SafePolynomial& operator-=(SafeCoeff a)					{ return ((*this) -= SafePolynomial(0,a)); }
	
	bool isZero() const { return Degree() == ZERO_DEGREE; }
	int Degree() const { return coefficients.size() - 1; }
		
	string toString(const string &tag = "q", int power_shift = 0) const;
	void PrintCoefficients(FILE *output, int i, int j) const;
	void Divide_By_Two();
	std::vector<SafePolynomial> GetNonzeroCoefficients() const;
		
	static void AnnounceErrorMessage(const string &);
	static void AnnounceOverflowError(const SafeCoeff &, const SafeCoeff &, char);
	static void ReadFilePolynomials(const string &, std::set<std::pair<int,int> > &, std::map<std::pair<int,int>, SafePolynomial> &);
	static void WriteSafePolynomialArray(FILE *, int, SafePolynomial**);
	static void Delete_2D_PolynomialArray(SafePolynomial **, int);
	
	static bool ReadSinglePolynomial(ifstream &, int &, int &, SafePolynomial &);

	static SafePolynomial GetZero()	{ return SafePolynomial(); }
	static SafePolynomial MultiplyHalfPolynomials(const SafePolynomial &, const SafePolynomial &);	
	static SafePolynomial** ReadSafePolynomialArray(const string &, int);
	static SafePolynomial** InitializeSafePolynomialArray(int); 
	
};



// Method implements command line printing for SafePolynomials
ostream& operator<<(ostream& out, const SafePolynomial& p);



/*
 * Function: PrintList
 * ---------------------------------------------------------
 * Prints the list of coefficients of a polynomial indexed 
 * by i,j to an output text file. The printed output has the 
 * form
 *
 *   i,j;a,b,c,d,...
 *
 * where the polynomial {a,b,c,d,...} is the list of coefficients.
 * In the output the letters are integers, and 0 is represented 
 * The tag input is a string indicated the data type of T to 
 * be used in fprintf, e..g., "%ld"
 *
 */
template <class T>
void PrintList(FILE *output, int i, int j, const std::vector<T> &pol, const string& tag)
{
	fprintf(output,"%d,%d;",i,j);
	
	if(pol.size() == 0) fprintf(output,"0\n");
	else {
		
		// Prints all but top degree coefficient, followed by commas
		// 0 coefficients are represented by empty string ""
		for(int d=0; d<pol.size()-1; d++) 
		{
			if(pol[d] == 0) fprintf(output, ",");
			else fprintf(output, (tag + ",").c_str(), pol[d]);
		}
		
		// Prints top degree coefficient, followed by newline
		fprintf(output, (tag + "\n").c_str(), pol.back());
	}
}



/*
 * Function: PrintPolynomial
 * ---------------------------------------------------------
 * Calls PrintList on the vector of coefficients of an arbitrary 
 * polynomial, as defined in "Polynomials.h"
 *
 */
template <class T>
void PrintPolynomial(FILE *output, int i, int j, const Polynomial<T> &pol)
{
	if(pol.isZero()) return;
	
	std::vector<T> vec;
	for(int n=0; n<=pol.deg(); n++) vec.push_back(pol[n]);
	
	PrintList(output, i, j, vec, SafeCoeff_TAG);
}



#endif