
/* 
 * File: "CoxeterExtensions_by_EM_Structs.cpp"
 *
 * Author: Eric Marberg
 * Version: June 2013
 * --------------------------------------------------------------
 * Source file for "CoxeterExtensions_by_EM_Structs.h"
 *
 */




#include "CoxeterExtensions_by_EM_Structs.h"




/*
 * Function: CheckCoefficients_ZeroInOddDegrees
 * ---------------------------------------------------------
 * Returns true if input polynomial has nonzero coefficients
 * only for monomials x^{2i} with even degree.
 *
 */
bool CheckCoefficients_ZeroInOddDegrees(const SafePolynomial &p)
{	
	for(int i=0; i<=p.Degree(); i++) 
	{
		if(p[i] != 0 && (i % 2) != 0) return false;
	}
	
	return true;	
}



/*
 * Function: CheckCoefficients_Nonnegative
 * ---------------------------------------------------------
 * Returns true if all coefficinets of the input polynomial
 * are nonnegative integers, false otherwise.
 *
 */
bool CheckCoefficients_Nonnegative(const SafePolynomial &p)
{	
	for(int i=0; i<=p.Degree(); i++) if(p[i] < 0) return false;
	return true;	
}



/*
 * Function: CheckCoefficients_Logconcave_in_q
 * ---------------------------------------------------------
 * Returns true if the input polynomial p is log-concave, false otherwise.
 * A polynomial is log-concave if whenever d,e,f are consecutive nonzero 
 * coefficients (of monomials of increasing degree), one has e^2 >= d*f.
 * A log-concave polynomial is automatically unimodal.
 *
 */
bool CheckCoefficients_Logconcave(const SafePolynomial &p)
{	
	if(! CheckCoefficients_ZeroInOddDegrees(p)) return false;
	
	for(int i = 1; i+1 < p.Degree()/2; i++)
	{
		SafeCoeff a = p[2*i-2];
		SafeCoeff b = p[2*i];
		SafeCoeff c = p[2*i+2];
		
		if( (a <= 0 && c >= 0) || (a >= 0 && c <= 0) ) continue;
		
		if( b == 0 ) return false; 
		if( 2*log(llabs(b)) < log(llabs(a)) + log(llabs(c)) ) return false;
	}
	
	return true;	
}



/*
 * Function: CheckCoefficients_Unimodal_in_q
 * ---------------------------------------------------------
 * Returns true if the input polynomial p is unimodal as a polynomial
 * in q = v^2, false otherwise.
 *
 * A polynomial a_0 + a_1q + ... + a_nq^n is unimodal if we have 
 *
 *   a_0 <= ... <= a_{k-1} <= a_k >= a_{k+1} >= ... >= a_n
 *
 * for some k.
 */
bool CheckCoefficients_Unimodal_in_q(const SafePolynomial &p)
{	
	if(! CheckCoefficients_ZeroInOddDegrees(p)) return false;
	
	
	bool started_decreasing = false;
	
	for(int i = 1; i < p.Degree()/2; i++)
	{
		if(!started_decreasing)
		{
			started_decreasing = (p[2*i] < p[2*i-2]);
		}
		else
		{
			if(p[2*i-2] < p[2*i]) return false;
		}
	}
	
	return true;	
}



/*
 * Function: CheckPolynomials_Decreasing
 * ---------------------------------------------------------
 * The input parameter polyns stores a family of polynomials f_{y,w}
 * indxed by two twisted involutions y,w in I_*. This function returns
 * true if the difference
 *
 *   f_{y,w} - f_{z,w}
 *
 * has nonnegative coefficients whenever z covers y in the 
 * Bruhat order. (If this holds that the difference has nonnegative
 * coefficients more generally whenever y <= z.) 
 *
 */
bool CheckPolynomials_Decreasing(std::map<std::pair<int,int>, SafePolynomial> & polyns, const SafePolynomial &f, const std::vector<int> &z_indices, int w_index)
{
	for(int i = 0; i < z_indices.size(); i++)
	{
		std::pair<int,int> new_pair = make_pair(z_indices[i], w_index);	
		
		SafePolynomial g = f;
		g -= polyns[new_pair];

		if(! CheckCoefficients_Nonnegative(g)) return false;
	}	
	
	return true;
}



/*
 * Function: GetCoefficientStatistics
 * ---------------------------------------------------------
 * Iterates over all polynomials in distinct_list and records
 * the largest and smallest nonzero coefficients which occur
 * (as the fields maxcoeff and mincoeff), the polynomials 
 * containing these coefficients (as the fields maxcoeff_pol
 * and mincoeff_pol), and the indices of these polynomials 
 * (as the vectors maxcoeff_index and mincoeff_index).
 *
 */
void ComparisonData_t::GetCoefficientStatistics()
{
	std::map<SafePolynomial, std::vector<int> >::iterator it;
	for(it = distinct_list.begin(); it != distinct_list.end(); ++it)
	{
		SafePolynomial f = it->first;
		std::vector<int> index = it->second;
		
		std::vector<SafePolynomial> nonzero_coeffs = f.GetNonzeroCoefficients();
		for(int i=0; i<nonzero_coeffs.size(); i++)
		{
			SafeCoeff c = nonzero_coeffs[i][0];

			if(c > maxcoeff) 
			{
				maxcoeff = c;
				maxcoeff_index = index;
				maxcoeff_pol = f;
			}
			
			if(c < mincoeff) 
			{
				mincoeff = c;
				mincoeff_index = index;
				mincoeff_pol = f;
			}
		}
	}
}



/*
 * Function: UpdatePolynomials
 * ---------------------------------------------------------
 * This function adds all polynomials stored (as values) in
 * the map tmp_list to the map distinct_list (as keys), then
 * clears tmp_list.
 *
 * The function takes as input the array whose yth entry is the 
 * list of (the indices of) the twisted involutions which cover
 * y in the Bruhat order on I_*. This is used to check that 
 * family of polynomials f_{y,w} stored in tmp_list is decreasing
 * with respect to the Bruhat order on y (when w is fixed).
 *
 */								  
void ComparisonData_t::UpdatePolynomials(std::vector<int>* bruhat_ascents)
{
	SafePolynomial f;
	int y, w;
	
	std::map<std::pair<int,int>, SafePolynomial>::iterator it;
	for(it = tmp_list.begin(); it != tmp_list.end(); ++it)
	{
		f = it->second;
		y = (it->first).first;
		w = (it->first).second;
		
		// Checks that f_{y,w} - f_{z,w} has nonnegative coefficients whenever y <= z
		polyns_decreasing &= CheckPolynomials_Decreasing(tmp_list, f, bruhat_ascents[y], w);
		
		// Creates value for f in distinct_list
		std::vector<int> index;
		index.push_back(y);
		index.push_back(w);
		
		// Adds f to distinct_list (keys of this map are distinct polynomials f)
		distinct_list[f] = index;
	}

	GetCoefficientStatistics();
	
	tmp_list.clear();
}



/*
 * Function: UpdateStructureConstants
 * ---------------------------------------------------------
 * This function adds all polynomials stored (as values) in
 * the map tmp_list to the map distinct_list (as keys), then
 * clears tmp_list.
 *
 * The function takes as input an integer y, which indicates
 * the polynomial stored in tmp_list as the value of the key 
 * (x,z) should be given the triple index (x,y,z).
 *
 */
void ComparisonData_t::UpdateStructureConstants(int y)
{
	SafePolynomial f;
	int x, z;
	
	distinct_list.clear();
	
	std::map<std::pair<int,int>, SafePolynomial>::iterator it;
	for(it = tmp_list.begin(); it != tmp_list.end(); ++it)
	{
		f = it->second;
		x = (it->first).first;
		z = (it->first).second;
		
		// Creates value for f in distinct_list
		std::vector<int> index;
		index.push_back(x);
		index.push_back(y);
		index.push_back(z);
		
		// Adds f to distinct_list
		distinct_list[f] = index;
	}
	
	GetCoefficientStatistics();
	CheckProperties_of_StructureConstants();
	
	tmp_list.clear();
}



/*
 * Function: CountProperty
 * ---------------------------------------------------------
 * The functions takes as input a pointer to a function with 
 * a single SafePolynomial argument and with a bool return value. 
 * The input function test is called on all polynomial keys
 * f in distinct_list, and the number of times that test(f) is 
 * true is returned as a long.
 * 
 */
long ComparisonData_t::CountProperty(bool(*test)(const SafePolynomial &))
{
	long answer = 0;
	
	std::map<SafePolynomial, std::vector<int> >::iterator it;
	for(it = distinct_list.begin(); it != distinct_list.end(); ++it)
	{
		if(test( it->first )) answer++;
	}
	
	return answer;
}



/*
 * Function: InterpretCount
 * ---------------------------------------------------------
 * If x == total, returns the string "true". 
 * If x != total, returns the string "false (true for 'x' of 'total')".
 *
 */
string InterpretCount(int x, int total) 
{
	if(x == total) return "true";
	
	stringstream ss;
	ss << "false (true for " << x << " of " << total << ")";
	return ss.str();
}



/*
 * Function: PrintProperty
 * ---------------------------------------------------------
 * This function takes three parameters: a stringstream, a string
 * label, and a function pointer. The function pointer represents
 * a boolean function which tests some property of a SafePolynomial.
 * The function pointer is applied to all polynomials in distinct_list,
 * and the results are printed to the stringstream, using the label.
 *
 */
void ComparisonData_t::PrintProperty(stringstream &ss, const string &label, bool(*test)(const SafePolynomial &))
{
	ss << label << InterpretCount(CountProperty(test), size()) << endl;
}

// Variant of the previous function where a bool is given in place of a function pointer
void ComparisonData_t::PrintProperty(stringstream &ss, const string &label, bool b)
{
	ss << label << (b ? "true" : "false") << endl;
}



/*
 * Function: CheckProperty
 * ---------------------------------------------------------
 * This function takes two parameters: a string label and a
 * function pointer. The function pointer represents a boolean 
 * function which tests some property of a SafePolynomial.
 * The function pointer is applied to all polynomials in distinct_list.
 * If the function pointer ever returns false, we set the value 
 * in the map "properties" under the key "tag" to be false. 
 * 
 */
void ComparisonData_t::CheckProperty(const string &tag, bool(*test)(const SafePolynomial &))
{
	// If we have not seen this property tag previously, 
	// add it to the list properties
	//
	if(properties.find(tag) == properties.end()) 
	{
		property_tags.push_back(tag);
		properties[tag] = true;
	}
	else
	{
		if(! properties[tag]) return;
	}
	
	// Iterates over all polynomials currently in distinct_list,
	// setting the properties[tag] to false if the function test
	// ever returns false.
	//
	std::map<SafePolynomial, std::vector<int> >::iterator it;
	for(it = distinct_list.begin(); it != distinct_list.end(); ++it)
	{
		if(! test( it->first )) 
		{
			properties[tag] = false;
			return;
		}
	}
	
}



/*
 * Function: PrintPropertiesHeader
 * ---------------------------------------------------------
 * Helper method called by PrintProperties which constucts
 * some explanatory text and appends it the input stringstream.
 *
 */
void ComparisonData_t::PrintPropertiesHeader(stringstream &ss)
{
	stringstream indexing;
	indexing << " (indexed by ";
	
	if(isPrintingStructureConstants())
	{
		indexing << symbols[0] << " in W and " << symbols[1] << "," << symbols[2] << " in I_*)";
	}
	else
	{
		indexing << symbols[0] << "," << symbols[1] << " in I_*)";
	}
	
	//
	// At this point index.str() should be equal to either 
	// "(indexed by y,w in I_*)" or "(index by x in W and y,z in I_*)"
	//
	
	ss << "Properties of the polynomials " << tag << indexing.str() << endl;
	ss << "------------------------------";
	
	for(int i=0; i<tag.size() + indexing.str().size(); i++) ss << "-";
	
	ss << endl << endl;
}



/*
 * Function: PrintProperties
 * ---------------------------------------------------------
 * This function prints to the input stringstream whether the 
 * polynomials stored in distinct_list satisfy a sequence of 
 * boolean properties. 
 *
 */
void ComparisonData_t::PrintProperties(stringstream &ss)
{
	PrintPropertiesHeader(ss);
	
	if(isPrintingStructureConstants()) 
	{
		PrintProperties_of_StructureConstants(ss);
	}
	else 
	{
		PrintProperties_of_Polynomials(ss);
	}
}



/*
 * Function: PrintProperties_of_Polynomials
 * ---------------------------------------------------------
 * Helper function called by PrintProperties if the family of
 * polynomials stored in the current ComparisonData_t is 
 * parametrized by three indices (e.g., h^{+/-}_{x,y;z}).
 *
 * To add more properties to be checked here, one can insert
 * additional calls to PrintProperty with a relevant string 
 * label and function pointer.
 *
 */
void ComparisonData_t::PrintProperties_of_Polynomials(stringstream &ss)
{	
	PrintProperty(ss,"Decreasing with respect to y index: ", polyns_decreasing);
	PrintProperty(ss,"Nonnegative coefficients          : ", CheckCoefficients_Nonnegative);
}



/*
 * Function: CheckProperties_of_StructureConstants
 * ---------------------------------------------------------
 * Helper function called by UpdateStructureConstants if the family of
 * polynomials stored in the current ComparisonData_t is 
 * parametrized by three indices (e.g., h^{+/-}_{x,y;z}).
 *
 * To add more properties to be checked here, one can insert
 * additional calls to CheckProperty with a relevant string 
 * label and function pointer.
 *
 */
void ComparisonData_t::CheckProperties_of_StructureConstants()
{
	CheckProperty("Polynomial in q when multiplied by v^degree : ", CheckCoefficients_ZeroInOddDegrees);
	CheckProperty("Nonnegative coefficients                    : ", CheckCoefficients_Nonnegative);
	CheckProperty("Unimodal in q when multiplied by v^degree   : ", CheckCoefficients_Unimodal_in_q);
	CheckProperty("Log-concave in q when multiplied by v^degree: ", CheckCoefficients_Logconcave);
}



/*
 * Function: PrintProperties_of_StructureConstants
 * ---------------------------------------------------------
 * Helper function called by PrintProperties if the family of
 * polynomials stored in the current ComparisonData_t is 
 * parametrized by three indices (e.g., h^{+/-}_{x,y;z}).
 *
 */
void ComparisonData_t::PrintProperties_of_StructureConstants(stringstream &ss)
{
	for(int i=0; i<property_tags.size(); i++)
	{
		ss << property_tags[i] << (properties[property_tags[i]] ? "true" : "false") << endl;
	}
}





/*
 * Function: Setup
 * ---------------------------------------------------------
 * The functions reads the list of polynomials in the given input file
 * and sets up the parameter z_lists. Each line of the input has the
 * form "y,z;a,b,c..." and z_lists is a vector whose y^th entry is the
 * array of z's for which a line of the input file begins "y,z;..."
 *
 */
void TransposeData_t::Setup(const string & inputfile) 
{
	SafePolynomial::ReadFilePolynomials( inputfile, yz_pairs, polynomials );
	
	
	// The vector z_lists is constructed below so that z_lists[y] is the set of 
	// indices z such that the pair (y,z) is a key in the map polynomials.
	
	std::set<std::pair<int,int > >::const_iterator yz_it;
	for(yz_it = yz_pairs.begin(); yz_it != yz_pairs.end(); ++yz_it)
	{
		z_lists[yz_it->first].push_back(yz_it->second);
	}
}
	


/*
 * Function: PrintEntries
 * ---------------------------------------------------------
 * Prints to a text file all f_{x,y;z} in the map polynomials 
 * for a fixed pair (x,y). Each line of the output is labeled 
 * by the indices x and z; the index y is stored as part of 
 * the name of the output file.
 *
 */
void TransposeData_t::PrintEntries(FILE* file, int x, int y)
{
	for(int i = 0; i < z_lists[y].size(); i++)
	{
		int z = z_lists[y][i];
		polynomials[ make_pair(y,z) ].PrintCoefficients(file, x, z);	
	}
		
	fclose(file);		
	z_lists[y].clear();
}
	

	
/*
 * Function: AllocateLists
 * ---------------------------------------------------------
 * Creates the lists d_edge and d_coeff to be of size l-by-d_size
 *
 */
void LowerBruhatIntervalsData_t::AllocateLists(const Ulong &d_size, const Rank& l)
{
	d_edge.setSize(l);
	d_coeff.setSize(l);
		
	for (Generator s = 0; s < l; ++s) 
	{
		(d_edge)[s].setSize(d_size);
		(d_coeff)[s].setSize(d_size);
	}
}		
	
	

/*
 * Function: FillLists
 * ---------------------------------------------------------
 * Fills d_edge so that d_edge[s][z] is the list of CoxNbr's x such that sx < x <= z.
 * Fills d_coeff so that d_coeff[s][z][i] is mu(P_{x,z}) hwere d_edge[s][z][i] = x.
 *
 * This code is copied verbatim from the GraphData constructor (around line 600 in "special.cpp").
 *
 */
void LowerBruhatIntervalsData_t::FillLists(const wgraph::WGraph &X)
{
	for (CoxNbr x = 0; x < X.size(); ++x) 
	{
		const wgraph::EdgeList& e = X.edge(x);
		const wgraph::CoeffList& c = X.coeffList(x);
		LFlags fx = X.descent(x);
			
		for (Ulong j = 0; j < e.size(); ++j) 
		{
			CoxNbr z = e[j];
			LFlags fz = X.descent(z);
				
			// look at the generators s.t. sx < x and sz > z
			for (LFlags f = fx & ~fz; f; f &= f-1) 
			{
				Generator s = firstBit(f);
				d_edge[s][z].append(x);
				d_coeff[s][z].append(c[j]);
			}
		}
	}
}		
	

